/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "abeditor.h"

#include <qapplication.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include "ablineedit.h"
#include "abmultilineedit.h"
#include "abentry.h"
#include "abname.h"

/*!
  \class AbEditor abeditor.h

  \brief AbEditor is a widget for editing an address book entry.
*/

AbEditor::AbEditor( AbEntry* entry, QWidget* parent = 0,
		    const char* name = 0, WFlags fl = 0 )
    : AbEditorBase( parent, name, fl ),
      nameDirty( FALSE )
{
    setEntry( entry );

    connect( FullNameLineEdit, SIGNAL( textChanged( const QString &) ),
	     this, SLOT( fullNameEdited() ) );
    connect( FullNameButton, SIGNAL( clicked() ),
	     this, SLOT( fullName() ) );
    connect( CompanyLineEdit, SIGNAL( textChanged( const QString &) ),
	     this, SLOT( companyEdited() ) );
    connect( CompanyButton, SIGNAL( clicked() ),
	     this, SLOT( company() ) );
}

AbEditor::~AbEditor()
{
}

void AbEditor::hideEvent( QHideEvent *e )
{
    QWidget *w = qApp->focusWidget();
    if ( w )
	w->clearFocus();
}

void AbEditor::setEntry( AbEntry* entry )
{
    ent = entry;

    FullNameButton->setFocus();
    FullNameLineEdit->init( ent, "FullName", tr( "< Full Name >"  ) );
    TitleLineEdit->init( ent, "Role", tr( "< Title >"  ) );
    DepartmentLineEdit->init( ent, "Department", tr( "< Department >"  ) );
    CompanyLineEdit->init( ent, "Company", tr( "< Company >"  ) );
    StreetLineEdit->init( ent, "AddressStreet", tr( "< Address, Street >"  ) );
    CityLineEdit->init( ent, "AddressCity", tr( "< Address, City >"  ) );
    StateLineEdit->init( ent, "AddressState", tr( "< Address, State/Province >"  ) );
    ZipLineEdit->init( ent, "AddressZip", tr( "< Address, Zip/Postal Code >"  ) );
    CountryLineEdit->init( ent, "AddressCountry", tr( "< Address, Country >"  ) );
    NotesMultiLineEdit->init( ent, "Notes", tr( "< Notes >"  ) );
    LineEdit1->init( ent, "0", "" );
    LineEdit2->init( ent, "1", "" );
    LineEdit3->init( ent, "2", "" );
    LineEdit4->init( ent, "3", "" );
    LineEdit5->init( ent, "4", "" );
    LineEdit6->init( ent, "5", "" );

    const uint numEdits = 6;
    int fo[numEdits] = { 4, 8, 3, 1, 0, 6 };

    const QString *fieldOrder = ent->find( "X-FieldOrder" );
    if (fieldOrder)
        sscanf( fieldOrder->local8Bit(), "%d,%d,%d,%d,%d,%d",
		&fo[0], &fo[1], &fo[2], &fo[3], &fo[4], &fo[5] );

    ComboBox1->setCurrentItem( fo[0] );
    ComboBox2->setCurrentItem( fo[1] );
    ComboBox3->setCurrentItem( fo[2] );
    ComboBox4->setCurrentItem( fo[3] );
    ComboBox5->setCurrentItem( fo[4] );
    ComboBox6->setCurrentItem( fo[5] );
}

QString AbEditor::fieldOrder()
{
    QString s;
    s.sprintf("%i,%i,%i,%i,%i,%i",
	      ComboBox1->currentItem(),
	      ComboBox2->currentItem(),
	      ComboBox3->currentItem(),
	      ComboBox4->currentItem(),
	      ComboBox5->currentItem(),
	      ComboBox6->currentItem() );

    return s;
}

void AbEditor::parseName()
{
    //TODO Check to see if name dialog has been used.
    //TODO Handle parsing of "surname, firstname" format.
    QString name;
    QString prefix;
    QString suffix;
    QString first;
    QString middle;
    QString last;
    if ( ent->find( "FullName" ) )
	name = *ent->find( "FullName" );

    name = name.simplifyWhiteSpace();
    if ( name.find( tr( "the" ), 0, FALSE ) != 0 ) {
	QString sTitle[] = {
	    tr( "Doctor" ), tr( "Dr." ), tr( "Dr" ), tr( "Miss" ),
	    tr ( "Mr." ), tr( "Mr" ), tr( "Mrs." ), tr( "Mrs" ),
	    tr( "Ms." ), tr( "Ms" ), tr( "Professor" ), tr( "Prof." ),
	    ""
	};
	QString sSuffix[] = {
	    tr( "IIII" ), tr( "II" ), tr( "I" ), tr( "Junior" ),
	    tr( "Jr." ), tr( "Senior" ), tr( "Sr." ),
	    ""
	};

	for ( int i =0; sTitle[i] != ""; ++i )
	    if ( name.find( sTitle[i], 0, FALSE ) == 0 ) {
		prefix = sTitle[i];
		name = name.right( name.length() - prefix.length() - 1 );
		name = name.simplifyWhiteSpace();
		break;
	    }

	for ( int i =0; sSuffix[i] != ""; ++i ) {
	    QString tSuffix = sSuffix[i];
	    int pos = name.length() - tSuffix.length();
	    if ( (pos > 0) && (name.findRev( tSuffix, -1, FALSE ) == pos) ) {
		suffix = tSuffix;
		name = name.left( pos - 1 );
		name = name.simplifyWhiteSpace();
		break;
	    }
	}
	if ( name.find( " ", 0 ) > 0 ) {
	    int pos = name.findRev( " ", -1 );
	    last = name.mid( pos + 1 );
	    name = name.left( pos );
	    name = name.simplifyWhiteSpace();
	    if ( name.find( " ", 0 ) > 0 ) {
		pos = name.find( " ", 0 );
		first = name.left( pos );
		name = name.mid( pos + 1 );
		name = name.simplifyWhiteSpace();
		middle = name;
	    }
	    else
		first = name;
	}
	else
	    last = name;
    }
    ent->replace( "Title", new QString( prefix ) );
    ent->replace( "FirstName", new QString( first ) );
    ent->replace( "MiddleName", new QString( middle ) );
    ent->replace( "LastName", new QString( last ) );
    ent->replace( "Suffix", new QString( suffix ) );
}

void AbEditor::fullName()
{
    AbName nameDialog( ent, this, "name dialog", true );
    nameDialog.setMaximumHeight( nameDialog.minimumHeight() );
    nameDialog.showMaximized();
    if ( nameDialog.exec() ) {
	FullNameLineEdit->sync();
	// FullNameLineEdit->setText( newName.simplifyWhiteSpace() );
	nameDirty = FALSE;
    }
}

void AbEditor::fullNameEdited()
{
    nameDirty = TRUE;
}

void AbEditor::company()
{
}

void AbEditor::companyEdited()
{
    companyDirty = TRUE;
}
