/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "abentry.h"
#include <qdict.h>
#include <qfile.h>
#include <qregexp.h>

////////////////////////////
// AbEntryList methods

AbEntryList::AbEntryList( const QString &filename )
{
    kkey = 0;
    setAutoDelete( true );
    load( filename );
}

QString AbEntryList::key()
{
    ++kkey;
    return QString().setNum( kkey );
}

QString AbEntryList::insert( AbEntry *item )
{
    QString result = key();
    QDict<AbEntry>::insert( result, item );
    return result;
}

void AbEntryList::unremove( const QString &key, AbEntry *item )
{
    QDict<AbEntry>::insert( key, item );
}

void AbEntryList::save( const QString &filename )
{
    QFile f( filename );
    if ( !f.open( IO_WriteOnly ) )
	return;

    QTextStream t( &f );

    QDictIterator<AbEntry> it(*this);
    while ( it.current() ) {
	it.current()->save( t );
	++it;
    }

    f.close();
}

void AbEntryList::load( const QString &filename )
{
    QFile f( filename );
    if ( !f.open( IO_ReadOnly ) )
	return;

    QTextStream t( &f );
    clear();

    while ( !t.eof() )
	QDict<AbEntry>::insert( key(), new AbEntry( t ));

    f.close();
}


////////////////////////
// AbEntry methods

AbEntry::AbEntry()
{
    dict.setAutoDelete( true );
}

AbEntry::AbEntry( const AbEntry &r )
  : QObject()
{
    QDictIterator<QString> it( r.dict );

    while ( it.current() ) {
	dict.replace( it.currentKey(), new QString( *it.current() ));
	++it;
    }
}

AbEntry& AbEntry::operator=( const AbEntry &r )
{
    if ( this != &r ) {
	dict.clear();
	QDictIterator<QString> it( r.dict );

	while ( it.current() ) {
	    dict.replace( it.currentKey(), new QString( *it.current() ));
	    ++it;
	}
    }
    return *this;
}

AbEntry::AbEntry( QTextStream &t )
{
    dict.setAutoDelete( true );
    load( t );
}

QStringList AbEntry::custom() const
{
    QStringList result;
    QDictIterator<QString> it( dict );

    while ( it.current() ) {
	if (it.currentKey().find( "X-CUSTOM-", 0 ) == 0)
	    result << it.currentKey();
	++it;
    }
    return result;
}

void AbEntry::save( QTextStream &t )
{
    QDictIterator<QString> it( dict );
    QRegExp reg("\n");

    while ( it.current() ) {
	if ( (it.currentKey().find( ".", 0 ) != 0) &&
	     (!(*it.current()).isEmpty()) ) {
	  t << " " << it.currentKey() << "\n";
	    QString tmp = *it.current();
	    tmp.replace( reg, "\n " );
	    t << " " << tmp << "\n[EOR]\n";
	}
	++it;
    }
    t << "[EOS]\n";
}

void AbEntry::load( QTextStream &t )
{
    while ( !t.eof() ) {
	QString name = t.readLine();
	if ( name == "[EOS]" )
	    break;
	name = name.mid(1);
	QString tmp = t.readLine();
	QString value = "";
	while ( tmp != QString( "[EOR]" )) {
	    if ( !value.isEmpty() )
		value += "\n";
	    value += tmp.mid(1);
	    tmp = t.readLine();
	}
	if ( (name != "") && (value != "") )
	    dict.replace( name, new QString( value ));
    }
    emit changed();
}

void AbEntry::insert( const QString key, const QString *item )
{
    if ( item && (*item == "") )
	return;
    dict.insert( key, item );
    emit changed();
}

void AbEntry::replace( const QString key, const QString *item )
{
    QString *current = dict.find( key );
    if ( item ) {
	if ( current ) {
	    if ( *item != *current ) {
		if ( *item == "" )
		    dict.remove( key ); // temporary?
		else
		    dict.replace( key, item );
		emit changed();
	    }
	}
	else { // (item && !current)
	    dict.replace( key, item );
	    emit changed();
	}
    }
    else
	debug( "Error:"
	       " AbEntry::replace( const QString, const QString* ) "
	       "passed null item" );
}

bool AbEntry::remove( const QString key )
{
    if ( dict.remove( key ) ) {
	emit changed();
	return TRUE;
    } else {
	return FALSE;
    }
}

void AbEntry::touch()
{
    emit changed();
}

const QString *AbEntry::find ( const QString & key ) const
{
    return dict.find( key );
}

const QString *AbEntry::operator[] ( const QString & key ) const
{
    return dict[key];
}

void AbEntry::clear ()
{
    dict.clear();
    emit changed();
}
