/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "abtable.h"
#include "abentry.h"


/*!
  \class AbTableItem abtable.h

  \brief QTableItem based class for showing a field of an entry
*/

AbTableItem::AbTableItem( QTable *t, EditType et, const QString &s,
			  const QString &k, int alignment )
    : QTableItem( t, et, s ), align( alignment )
{
    entKey = k;
}

QString AbTableItem::entryKey() const
{
    return entKey;
}

void AbTableItem::setEntryKey( const QString & k )
{
    entKey = k;
}

int AbTableItem::alignment() const
{
    return align;
}

QString AbTableItem::key() const
{
    return text().lower();
}

/*!
  \class AbTable abtable.h

  \brief QTable based class for showing a list of entries
*/

AbTable::AbTable( QWidget *parent, const char *name )
    : QTable( parent, name ), lastSortCol( -1 ), asc( TRUE )
{
    setSorting( TRUE ); //TODO get sorting working
    setSelectionMode( NoSelection );
}

AbTable::~AbTable()
{
}

void AbTable::init( AbEntryList *iCel )
{
    QDictIterator<AbEntry> it( *iCel );
    cel = iCel;

    setNumRows( 0 );
    setNumCols( 2 );

    horizontalHeader()->setLabel( 0, tr( "Full Name" ));
    horizontalHeader()->setLabel( 1, tr( "Contact" ));
    setLeftMargin( 0 );
    verticalHeader()->hide();
    /*
      setTopMargin( 0 );
      horizontalHeader()->hide();
    */

    QDictIterator<AbEntry> jt( *cel );
    if ( !jt.current() )
	emit empty( TRUE );	
    while ( jt.current() ) {
        insertEntry( jt.current(), jt.currentKey() );
        ++jt;
    }
}

void AbTable::setCell( int row, int col, const QString &value,
		       const QString &key )
{
    int align = AlignLeft | AlignVCenter;
    if (col == 1)
	align = AlignRight | AlignVCenter;
    QTableItem *ati;
    delete item( row, col );
    ati = new AbTableItem( this, QTableItem::WhenCurrent, value, key, align );
    setItem( row, col, ati );
}

void AbTable::setRow( int row, AbEntry *entry, QString key )
{
    QString firstName, lastName, name;
    QString fieldNum, indicator;
    const int numFields = 6;
    const int numTypes = 8;
    int firstField;
    const QString *value;

    if ( entry->find( "LastName" ) )
	lastName = *entry->find( "LastName" );
    if ( entry->find( "FirstName" ) )
	firstName = *entry->find( "FirstName" );
    if (!lastName.isEmpty() && !firstName.isEmpty())
	name = " " + lastName + ", " + firstName;
    else
	name = " " + firstName + lastName;
    //    const QString *value = entry->find( "FullName" );
    setCell( row, 0, name, key );
	
    for ( firstField = 0; firstField < numFields; ++firstField ) {
	fieldNum.setNum( firstField );
	if ( entry->find( fieldNum ) )
	    break;
	}	
	
    const QString *fieldOrder = entry->find( "X-FieldOrder" );
    if ( fieldOrder ) {
	int f[numFields];
	sscanf( fieldOrder->local8Bit(), "%d,%d,%d,%d,%d,%d",
		&f[0], &f[1], &f[2], &f[3], &f[4], &f[5] );

	// How do I i18n single letters?
	switch ( QMIN( QMAX( 0, f[firstField] ), numTypes )) {
	    case 0:
		indicator = " B ";
		break;
	    case 1:
		indicator = " E ";
		break;
	    case 2:
		indicator = " F ";
		break;
	    case 3:
		indicator = " H ";
		break;
	    case 4:
		indicator = " M ";
		break;
	    case 5:
		indicator = " O ";
		break;
	    case 6:
		indicator = " S ";
		break;
	    case 7:
		indicator = " P ";
		break;
	    case 8:
		indicator = " W ";
		break;
	    default:
		indicator = "   ";
		break;
	    }
    }
	
    value = entry->find( fieldNum );
    if ( value )
        setCell( row, 1, *value + indicator, key );
    else
        setCell( row, 1, "", key );
}

AbEntry* AbTable::insertEntry( AbEntry *entry, QString key )
{
    setNumRows( numRows() + 1 );
    setRow( numRows() - 1, entry, key );
    if (numRows() == 1)
	emit empty( FALSE );
    return 0;
}

void AbTable::columnClicked( int col )
{
    if ( !sorting() )
	return;

    if ( col == lastSortCol ) {
	asc = !asc;
    } else {
	lastSortCol = col;
	asc = TRUE;
    }

    sortColumn( lastSortCol, asc, TRUE );
}

AbEntry* AbTable::currentEntry()
{
    AbTableItem *abItem = (AbTableItem*)item( currentRow(), 0 );
    if (!abItem)
	return 0;

    QString key = abItem->entryKey();
    return cel->find( key );
}

void AbTable::replaceCurrentEntry( AbEntry* newEntry )
{
    AbEntry *entry = currentEntry();
    ASSERT(entry);

    endEdit( currentRow(), currentColumn(), FALSE, FALSE );
    QTableItem *ti = item( currentRow(), 0 );
    ASSERT( ti );
    AbTableItem *ati = static_cast< AbTableItem* >(ti);
    QString key = ati->entryKey();
    cel->replace( key, newEntry );
    setRow( currentRow(), newEntry, key );
}

void AbTable::deleteCurrentEntry()
{
    endEdit( currentRow(), currentColumn(), FALSE, FALSE );
    if ( currentRow() >= 0 ) {
	AbTableItem *abItem = (AbTableItem*)item( currentRow(), 0 );
	if (!abItem)
	    return;
	QString key = abItem->entryKey();
        cel->remove( key );
        for ( int i = currentRow(); i < numRows() - 1; ++i ) {
	    abItem = (AbTableItem*)item( i + 1, 0 );
	    key = abItem->entryKey();
	    AbEntry *entry = cel->find( key );
	    ASSERT( entry );
	    setRow( i, entry, key );
        }
        setNumRows( numRows() - 1 ); //QGVector::remove: Index 4 out of range
    }

    if ( numRows() == 0 )
	emit empty( TRUE );
}

QWidget *AbTable::createEditor( int, int, bool ) const
{
    return 0;
}

