/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "addressbook.h"

#include <qimage.h>
#include <qpixmap.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qapplication.h>
#include <qwidgetstack.h>
#include <qdialog.h>
#include <qlayout.h>
#include "pixmaploader.h"

#include "abentry.h"
#include "abeditor.h"
#include "ablabel.h"
#include "abtable.h"

#include "new.xpm"
#include "zoom.xpm"
#include "trash.xpm"
#include "back.xpm"
#include "pencil.xpm"

#include <stdlib.h>

static QString addressbookFilename()
{
    QString filename=getenv("HOME");
    filename += "/Documents/addressbook";
    return filename;
}


AddressbookWindow::AddressbookWindow( QWidget *parent, const char *name, WFlags f )
    : QMainWindow( parent, name, f )
{
    setCaption( tr("Address Book") );

    setToolBarsMovable( FALSE );

    // Create Toolbars
    QPixmap newIcon, viewIcon, deleteIcon;
    QPixmap backIcon, editIcon;
    QPixmap okIcon, cancelIcon;

    listTools = new QToolBar( this, "list operations" );
    listTools->setLabel( tr( "List Operations" ) );

    newIcon = QPixmap( new_xpm );
    new QToolButton( newIcon, "New Entry", QString::null,
		     this, SLOT(slotListNew()), listTools, "edit entry" );

    viewIcon = QPixmap( zoom_xpm );
    viewButton = new QToolButton( viewIcon, "View Entry", QString::null,
		     this, SLOT(slotListView()), listTools, "edit entry" );

    deleteIcon = QPixmap( trash_xpm );
    deleteButton = new QToolButton( deleteIcon, "Delete Entry", QString::null,
		     this, SLOT(slotListDelete()), listTools, "delete entry" );

    viewTools = new QToolBar( this, "view operations" );
    viewTools->setLabel( tr( "View Operations" ) );

    backIcon = QPixmap( back_xpm );
    new QToolButton( backIcon, "Back", QString::null,
		     this, SLOT(slotViewBack()), viewTools, "back" );

    editIcon = QPixmap( pencil_xpm );
    new QToolButton( editIcon, "Edit Entry", QString::null,
		     this, SLOT(slotViewEdit()), viewTools, "edit entry" );

    viewTools->hide();

    // Create Views
    cel = new AbEntryList( addressbookFilename() );
    entry = new AbEntry();

    abList = new AbTable( this, "table" );
    abList->setHScrollBarMode( QScrollView::AlwaysOff );
    connect( abList, SIGNAL( empty( bool ) ),
	     this, SLOT( listIsEmpty( bool ) ) );
    abList->init( cel );
    abList->sortColumn( 0, TRUE, TRUE );

    sv2 = new QScrollView( this, "scrollview" );
    abView = new AbLabel( sv2->viewport(), "viewer" );
    abView->setAlignment( AlignLeft | AlignTop );
    viewMargin = 10;
    abView->setMargin( viewMargin );
    abView->init( entry  );
    sv2->addChild( abView );

    editDlg = new QDialog( this, 0, TRUE );
    editDlg->setCaption( tr("Edit Address") );
    QVBoxLayout *vb = new QVBoxLayout( editDlg );
    QScrollView *sv = new QScrollView( editDlg, "scrollview" );
    vb->addWidget( sv );
    abEditor = new AbEditor( entry, sv->viewport(), "editor" );
    sv->addChild( abEditor );

    sv2->hide();
    abView->hide();
    setCentralWidget( abList );
}

void AddressbookWindow::resizeEvent( QResizeEvent *e )
{
  QMainWindow::resizeEvent( e );

  if ( centralWidget() == abList  )
      showList();
  else if ( centralWidget() == sv2 )
      showView();
}

AddressbookWindow::~AddressbookWindow()
{
    cel->save( addressbookFilename() );
}

void AddressbookWindow::showList()
{
    viewTools->hide();
    listTools->show();
    abView->hide();
    sv2->hide();
    abList->show();

    setCentralWidget( abList );
    int contentsWidth = abList->visibleWidth();
    abList->setColumnWidth( 0, contentsWidth - contentsWidth / 2 );
    abList->setColumnWidth( 1, contentsWidth / 2 );
}

void AddressbookWindow::showView()
{
    listTools->hide();
    viewTools->show();
    abList->hide();
    abView->show();
    sv2->show();

    QSize oldSize = centralWidget()->size();
    sv2->resize( oldSize );
    int contentsWidth =  sv2->visibleWidth();

    if ( abView ) {
        abView->setMaximumWidth( contentsWidth );
	abView->setMinimumWidth( contentsWidth );
	abView->setMinimumHeight( sv2->visibleHeight() - 2 );
    }

    setCentralWidget( sv2 );
    sv2->updateGeometry();
}

void AddressbookWindow::slotListNew()
{
    delete entry;
    entry = new AbEntry();
    abEditor->setEntry( entry );
    abView->init( entry );
    entry->touch();
    editEntry( NewEntry );
}

void AddressbookWindow::slotListView()
{
    if ( !abList->currentEntry() )
      return;
    *entry = *(abList->currentEntry());
    abView->sync();
    showView();
}

void AddressbookWindow::slotListDelete()
{
    abList->deleteCurrentEntry();
    showList();
}

void AddressbookWindow::slotViewBack()
{
    showList();
}

void AddressbookWindow::slotViewEdit()
{
    abEditor->setEntry( entry );
    editEntry( EditEntry );
}

void AddressbookWindow::editEntry( EntryMode entryMode )
{
#ifdef _WS_QWS_
    editDlg->showMaximized();
#endif

    if ( editDlg->exec() == QDialog::Accepted ) {
	setFocus();
	abEditor->parseName();
	entry->replace( "X-FieldOrder", new QString( abEditor->fieldOrder() ));
	if ( entryMode == NewEntry ) {
	    AbEntry *insertEntry = new AbEntry( *entry );
	    QString key = cel->insert( insertEntry );
	    abList->insertEntry( insertEntry, key );
	} else
	    abList->replaceCurrentEntry( new AbEntry( *entry ));
    }
    showList();
}

void AddressbookWindow::listIsEmpty( bool empty )
{
    if ( empty ) {
	viewButton->setEnabled( FALSE );
	deleteButton->setEnabled( FALSE );
    } else {
	viewButton->setEnabled( TRUE );
	deleteButton->setEnabled( TRUE );
    }
}
