/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "datebookday.h"
#include "datebookdayheaderimpl.h"
#include <pixmaploader.h>

#include <qheader.h>
#include <qdatetime.h>
#include <qpainter.h>
#include <qsimplerichtext.h>
#include <qpopupmenu.h>
#include <qtextcodec.h>
#include <qpalette.h>

DateBookDayItem::DateBookDayItem( QTable *parent, int t )
    : QTableItem( parent, Never, QString::null )
{
    QString s = QString::number( t );
    if ( s.length() == 1 )
	s.prepend( "0" );
    setText( s );
}

DateBookDayView::DateBookDayView( QWidget *parent, const char *name )
    : QTable( 24, 1, parent, name )
{
    enableClipper(TRUE);
    setTopMargin( 0 );
    horizontalHeader()->hide();
    for ( int i = 0; i < 24; ++i ) {
	verticalHeader()->setLabel( i, QString::number(i)+":00" );
	(void)new DateBookDayItem( this, i );
	setRowStretchable( i, FALSE );
    }
    setLeftMargin(34);
    verticalHeader()->setPalette(white);
    verticalHeader()->setResizeEnabled(FALSE);
}

DateBookDay::DateBookDay( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    header = new DateBookDayHeader( this, "day header" );
    connect( header, SIGNAL( dateChanged( int, int, int ) ),
	     this, SLOT( dateChanged( int, int, int ) ) );
    header->setDate( QDate::currentDate().year(), QDate::currentDate().month(), QDate::currentDate().day() );
    view = new DateBookDayView( this, "day view" );
    currDate = QDate::currentDate();
}

void DateBookDay::saveAllEvents( const QString &fn )
{
    QFile f( fn );
    if ( !f.open( IO_WriteOnly ) )
        return;
    QTextStream ts( &f );
    ts.setCodec( QTextCodec::codecForName( "UTF-8" ) );
    ts << "<!DOCTYPE DATEBOOK><DATEBOOK>" << endl;

    for ( EventMap::Iterator it = events.begin(); it != events.end(); ++it ) {
	for ( DateBookDayWidget *w = (*it)->first(); w; w = (*it)->next() ) {
	    ts << "<event>" << endl;
	    w->event().save(ts," ");
	    ts << "</event>" << endl;
	}
    }

    ts << "</DATEBOOK>" << endl;
}

void DateBookDay::selectedDates( QDateTime &start, QDateTime &end )
{
    start.setDate( currDate );
    end.setDate( currDate );

    int sh=99,eh=-1;

    int n = dayView()->numSelections();

    if ( n ) {
	for (int i=0; i<n; i++) {
	    QTableSelection sel = dayView()->selection( i );
	    sh = QMIN(sh,sel.topRow());
	    eh = QMAX(sh,sel.bottomRow()+1);
	}
    } else {
	sh=1;
	eh=2;
    }

    start.setTime( QTime( sh, 0, 0 ) );
    end.setTime( QTime( eh, 0, 0 ) );
}

void DateBookDay::setDate( int y, int m, int d )
{
    header->setDate( y, m, d );
}

void DateBookDay::dateChanged( int y, int m, int d )
{
    QString s = currDate.toString();
    QMap<QString, EventList * >::Iterator it = events.find( s );
    if ( it != events.end() ) {
	for ( DateBookDayWidget *w = (*it)->first(); w; w = (*it)->next() )
	    w->hide();
    }
    currDate.setYMD( y, m, d );
    s = currDate.toString();
    it = events.find( s );
    if ( it != events.end() ) {
	for ( DateBookDayWidget *w = (*it)->first(); w; w = (*it)->next() )
	    w->show();
    }
}

void DateBookDay::addEvent( const Event &e )
{
    DateBookDayWidget *w = new DateBookDayWidget( e, this );
    QString s = currDate.toString();
    QMap<QString, EventList * >::Iterator it = events.find( s );
    if ( it == events.end() ) {
	events.insert( s, new EventList() );
	it = events.find( s );
	(*it)->setAutoDelete( TRUE );
    }
    (*it)->append( w );
    relayoutPage( *it );
    connect( w, SIGNAL( deleteMe( const QDate &, DateBookDayWidget * ) ),
	     this, SLOT( deleteEvent( const QDate &, DateBookDayWidget * ) ) );
    connect( w, SIGNAL( editMe( const QDate &, DateBookDayWidget * ) ),
	     this, SLOT( editEvent( const QDate &, DateBookDayWidget * ) ) );
    emit newEvent( e );
}

void DateBookDay::resizeEvent( QResizeEvent* e )
{
    QVBox::resizeEvent(e);
    QString s = currDate.toString();
    dayView()->setColumnWidth( 0, dayView()->visibleWidth() );
    QMap<QString, EventList * >::Iterator it = events.find( s );
    if ( it != events.end() )
	relayoutPage( *it );
}

void DateBookDay::relayoutPage( EventList *l )
{
    if ( !l || !l->count() )
	return;

    l->sort();

    for ( int i=0; i < (int)l->count(); ) {
	DateBookDayWidget *w = l->at(i);
	QTime et = w->event().end().time();
	int n = 1;
	for ( int j=i+1; j < (int)l->count(); j++ ) {
	    QTime jst = l->at(j)->event().start().time();
	    if ( jst >= et )
		break;
	    QTime jet = l->at(j)->event().end().time();
	    if ( jet > et )
		et = jet;
	    n++;
	}
	int wid = (view->columnWidth(0)-1) / n;
	for (int x=0; x<n; x++) {
	    w = l->at(i);
	    w->setX( x*wid );
	    w->setWidth( wid );
	    i++;
	}
    }
}

void DateBookDay::editEvent( const QDate &date, DateBookDayWidget *me )
{
    QString s = date.toString();
    QMap<QString, EventList * >::Iterator it = events.find( s );
    if ( it != events.end() ) {
	Event e = me->event();
	(*it)->removeRef( me );
	emit editEvent( e );
    }
}

void DateBookDay::deleteEvent( const QDate &date, DateBookDayWidget *me )
{
    QString s = date.toString();
    QMap<QString, EventList * >::Iterator it = events.find( s );
    if ( it != events.end() ) {
	Event e = me->event();
	(*it)->removeRef( me );
	relayoutPage( *it );
	emit removeEvent( e );
    }
}

void DateBookDay::deleteEvent( const Event &e )
{
    QMap<QString, EventList * >::Iterator it = events.begin();
    for ( ; it != events.end(); ++it ) {
	EventList *l = *it;
	for ( DateBookDayWidget *w = l->first(); w; w = l->next() ) {
	    if ( w->event() == e ) {
		l->removeRef( w );
		return;
	    }
	}
    }    
}

DateBookDayWidget::DateBookDayWidget( const Event &e, DateBookDay *db )
    : QWidget( db->dayView() ), ev( e ), dateBook( db ), xpos( 0 ), wid( 10 )
{
    place();
    text = "<b>" + ev.description() + "</b><br>" + "<i>" +
	   ev.category() + "</i><br><b>Start</b>: " + e.start().toString() +
	   "<br><b>End</b>: " + e.end().toString() + "<br><br>" + ev.notes();
    setBackgroundMode( PaletteBase );
}

void DateBookDayWidget::paintEvent( QPaintEvent *e )
{
    QPainter p( this );
    p.setPen(QColor(100,100,100));
    p.setBrush(QColor(255,240,230)); // based on priority?
    p.drawRect(rect());

    int y = 0;
    int d = 0;

    if ( ev.hasAlarm() ) {
	p.drawPixmap( width() - 16, 0, PixmapLoader::loadPixmap( "bell" ) );
	y = 20;
	d = 20;
    }

    if ( ev.doRepeat() ) {
	p.drawPixmap( width() - 16, y, PixmapLoader::loadPixmap( "repeat" ) );
	d = 20;
    }

    QSimpleRichText rt( text, font() );
    rt.setWidth( wid - d - 6 );
    rt.draw( &p, 3, 0, e->region(), colorGroup() );
}

void DateBookDayWidget::mousePressEvent( QMouseEvent *e )
{
    QPopupMenu m;
    m.insertItem( tr( "Edit..." ), 1 );
    m.insertItem( tr( "Delete" ), 2 );
    int r = m.exec( e->globalPos() );
    if ( r == 1 ) {
	emit editMe( ev.start().date(), this );
    } else if ( r == 2 ) {
	emit deleteMe( ev.start().date(), this );
    }
}

void DateBookDayWidget::place()
{
    QTime s = ev.start( TRUE ).time();
    QTime e = ev.end( TRUE ).time();
    int y = s.hour()*60+s.minute();
    int h = e.hour()*60+e.minute()-y;
    int rh = dateBook->dayView()->rowHeight(0);
    y = y*rh/60;
    h = h*rh/60;
    if ( h<3 ) h=3;
    setFixedSize( wid+1, h+1 );
    dateBook->dayView()->moveChild( this, xpos, y-1 );
    show();
}

void DateBookDayWidget::setX( int x )
{
    xpos = x;
    place();
}

void DateBookDayWidget::setWidth( int w )
{
    wid = w;
    place();
}

QDate DateBookDay::date() const
{
    return currDate;
}

int EventList::compareItems( QCollection::Item a, QCollection::Item b )
{
    DateBookDayWidget* aa = (DateBookDayWidget*)a;
    DateBookDayWidget* bb = (DateBookDayWidget*)b;
    return bb->event().start().secsTo(aa->event().start());
}
