/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "datebookweek.h"
#include "calendar.h"
#include "datebookweekheaderimpl.h"

#include <qheader.h>
#include <qdatetime.h>
#include <qpopupmenu.h>
#include <qtooltip.h>
#include <qpainter.h>

DateBookWeekItem::DateBookWeekItem( QListView *parent, int t )
    : QListViewItem( parent )
{
    QString s = QString::number( t );
    if ( s.length() == 1 )
	s.prepend( "0" );
    setText( 0, s );
}

void DateBookWeekItem::paintCell( QPainter *p, const QColorGroup & cg,
				 int column, int width, int alignment )
{
    QListViewItem::paintCell( p, cg, column, width, alignment );
    p->save();
    p->setPen( lightGray );
    p->drawLine( width - 1, 0, width - 1, height() );
    p->restore();
}

DateBookWeekView::DateBookWeekView( QWidget *parent, const char *name )
    : QListView( parent, name )
{
    addColumn( "" );
    addColumn( "M" );
    addColumn( "T" );
    addColumn( "W" );
    addColumn( "T" );
    addColumn( "F" );
    addColumn( "S" );
    addColumn( "S" );

    setSorting( -1 );

    for ( int i = 0; i < 24; ++i )
	(void)new DateBookWeekItem( this, 23 - i );

    header()->setResizeEnabled( FALSE );
    header()->setMovingEnabled( FALSE );
    setSelectionMode( NoSelection );
    connect( this, SIGNAL( clicked( QListViewItem *, const QPoint &, int ) ),
	     this, SLOT( dateClicked( QListViewItem *, const QPoint &, int ) ) );
}

void DateBookWeekView::resizeEvent( QResizeEvent *e )
{
    QListView::resizeEvent( e );
    QSize vs = viewportSize( 0, contentsHeight() );

    int ns = ( vs.width() - header()->sectionSize( 0 ) ) / 7;
    if ( ns < 16 )
	ns = 16;
	
    for ( int i = 1; i <= 7; ++i )
	header()->resizeSection( i, ns );
    header()->repaint();
    viewport()->repaint( FALSE );
}

void DateBookWeekView::dateClicked( QListViewItem *, const QPoint &, int c )
{
qDebug("%d",c);
    if ( c == 0 )
	return;
    emit showDay( c );
}


DateBookWeek::DateBookWeek( QWidget *parent, const char *name )
    : QVBox( parent, name )
{
    header = new DateBookWeekHeader( this );
    view = new DateBookWeekView( this );
    connect( view, SIGNAL( showDay( int ) ),
	     this, SLOT( showDay( int ) ) );
    connect( header, SIGNAL( dateChanged( int, int ) ),
	     this, SLOT( dateChanged( int, int ) ) );
    setDate( 2000, QDate::currentDate().dayOfYear() / 7 );
}

void DateBookWeek::showDay( int day )
{
    QDate d;
    d.setYMD( year, 1, 1 );
    d = d.addDays( ( week - 1 ) * 7 - d.dayOfWeek() + 1 );
    d = d.addDays( day - 1 );
    emit showDate( d );
}

void DateBookWeek::setDate( int y, int w )
{
    header->setDate( y, w );
}

void DateBookWeek::dateChanged( int y, int w )
{
    QString s = QString::number( year ) + "_" + QString::number( week );
    QMap<QString, QList<DateBookWeekWidget> * >::Iterator it = events.find( s );
    if ( it != events.end() ) {
	for ( DateBookWeekWidget *w = (*it)->first(); w; w = (*it)->next() )
	    w->hide();
    }
    year = y;
    week = w;
    s = QString::number( year ) + "_" + QString::number( week );
    it = events.find( s );
    if ( it != events.end() ) {
	for ( DateBookWeekWidget *w = (*it)->first(); w; w = (*it)->next() )
	    w->show();
    }
}

QDate DateBookWeek::date() const
{
    QDate d;
    d.setYMD( year, 1, 1 );
    d = d.addDays( ( week - 1 ) * 7 - d.dayOfWeek() + 1 );
    return d;
}

void DateBookWeek::addEvent( const Event &e )
{
    DateBookWeekWidget *w = new DateBookWeekWidget( e, this, e.start().date().year(), e.start().date().dayOfYear() / 52 );
    QString s = QString::number( e.start().date().year() ) + "_" + QString::number( e.start().date().dayOfYear() / 52 );
    QMap<QString, QList<DateBookWeekWidget> * >::Iterator it = events.find( s );
    if ( it == events.end() ) {
	events.insert( s, new QList<DateBookWeekWidget>() );
	it = events.find( s );
	(*it)->setAutoDelete( TRUE );
    }
    (*it)->append( w );
    relayoutPage( *it );
    connect( w, SIGNAL( deleteMe( int, int, DateBookWeekWidget * ) ),
	     this, SLOT( deleteEvent( int, int, DateBookWeekWidget * ) ) );
    emit newEvent( e );
}

void DateBookWeek::relayoutPage( QList<DateBookWeekWidget> *l )
{
    if ( !l || !l->count() )
	return;

    // ### need a bettrer implementation. At the moment each event
    // gets an own column. The real solution is to make as few columns
    // as possible, this means only make new columns if events overlap.
    int lastX[ 7 ];
    for ( int i = 0; i < 7; ++i )
	lastX[ i ] = 0;
    for ( DateBookWeekWidget *w = l->first(); w; w = l->next() ) {
	int dow = w->event().start().date().dayOfWeek();
	w->setX( view->header()->sectionPos( dow ) + lastX[ dow - 1 ] );
	w->setWidth( 2 );
	lastX[ dow - 1 ] += 2;
    }
}

void DateBookWeek::deleteEvent( int y, int w, DateBookWeekWidget *me )
{
    emit removeEvent( me->event() );
    QString s = QString::number( y ) + "_" + QString::number( w );
    QMap<QString, QList<DateBookWeekWidget> * >::Iterator it = events.find( s );
    if ( it != events.end() )
	(*it)->removeRef( me );
}

void DateBookWeek::deleteEvent( const Event &e )
{
    QMap<QString, QList<DateBookWeekWidget> * >::Iterator it = events.begin();
    for ( ; it != events.end(); ++it ) {
	QList<DateBookWeekWidget> *l = *it;
	for ( DateBookWeekWidget *w = l->first(); w; w = l->next() ) {
	    if ( w->event() == e ) {
		l->removeRef( w );
		return;
	    }
	}
    }    
}


DateBookWeekWidget::DateBookWeekWidget( const Event &e, DateBookWeek *db, int y, int w )
    : QWidget( db->weekView()->viewport() ), ev( e ), dateBook( db ),
      xpos( 30 ), wid( 100 ), year( y ), week( w )
{
    place();
    setBackgroundColor( e.hasAlarm() ? red : blue );
    QToolTip::add( this, "<b>" + ev.description() + "</b><br>" + "<i>" +
		   ev.category() + "</i><br><b>Start</b>: " + e.start().toString() +
		   "<br><b>End</b>: " + e.end().toString() + "<br><br>" + ev.notes() );
}

void DateBookWeekWidget::mousePressEvent( QMouseEvent *e )
{
    QPopupMenu m;
    m.insertItem( tr( "Edit..." ), 1 );
    m.insertItem( tr( "Delete" ), 2 );
    int r = m.exec( e->globalPos() );
    if ( r == 1 ) {
	// .... edit
    } else if ( r == 2 ) {
	emit deleteMe( year, week, this );
    }
}

void DateBookWeekWidget::place()
{
    show();
    int y = ev.start( TRUE ).time().hour();
    y *= dateBook->weekView()->firstChild()->height();
    int h = ev.end( TRUE ).time().hour() - ev.start().time().hour() + 1;
    h *= dateBook->weekView()->firstChild()->height();
    dateBook->weekView()->moveChild( this, xpos, y );
    resize( wid, h );
}

void DateBookWeekWidget::setX( int x )
{
    xpos = x;
    place();
}

void DateBookWeekWidget::setWidth( int w )
{
    wid = w;
    place();
}
