/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "dateentryimpl.h"
#include <datebookmonth.h>
#include <global.h>

#include <qcombobox.h>
#include <qtoolbutton.h>
#include <qpopupmenu.h>
#include <qlineedit.h>
#include <qcheckbox.h>
#include <qspinbox.h>
#include <qmultilineedit.h>

/*
 *  Constructs a DateEntry which is a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
DateEntry::DateEntry( const QDateTime &start, const QDateTime &end,
		      QWidget* parent,  const char* name )
    : DateEntryBase( parent, name, TRUE )
{
    init();
    setDates(start,end);
}

static void addOrPick( QComboBox* combo, const QString& t )
{
    for (int i=0; i<combo->count(); i++) {
	if ( combo->text(i) == t ) {
	    combo->setCurrentItem(i);
	    return;
	}
    }
    combo->setEditText(t);
}

DateEntry::DateEntry( const Event &event,
		      QWidget* parent,  const char* name )
    : DateEntryBase( parent, name, TRUE )
{
    init();
    setDates(event.start(),event.end());
    addOrPick( comboCategory, event.category() );
    addOrPick( comboDescription, event.description() );
    addOrPick( comboLocation, event.location() );
    checkAlarm->setChecked(event.hasAlarm());
    if ( event.type() == Event::AllDay )
	comboType->setCurrentItem(1);
    editNote->setText(event.notes());
    spinAlarm->setValue(event.alarmTime());
    checkRepeat->setChecked(event.doRepeat());
    //comboRepeat->
}

void DateEntry::setDates( const QDateTime& s, const QDateTime& e )
{
    startDate = s.date();
    endDate = e.date();
    startTime = s.time();
    endTime = e.time();
    startDateChanged( s.date().toString() );
    QString t;
    t.sprintf( "%02d:%02d", s.time().hour(), s.time().minute() );
    comboStart->setEditText( t );
    endDateChanged( e.date().toString() );
    t.sprintf( "%02d:%02d", e.time().hour(), e.time().minute() );
    comboEnd->setEditText( t );
}

void DateEntry::init()
{
    comboDescription->setInsertionPolicy(QComboBox::AtCurrent);
    comboLocation->setInsertionPolicy(QComboBox::AtCurrent);
    comboCategory->setInsertionPolicy(QComboBox::AtCurrent);

    DateBookMonth *picker = 0;
    QPopupMenu *m1 = new QPopupMenu( this );
    picker = new DateBookMonth( m1, 0, TRUE );
    m1->insertItem( picker );
    buttonStart->setPopup( m1 );
    buttonStart->setPopupDelay( 0 );
    startPicker = picker;
    connect( picker, SIGNAL( dateClicked( const QString & ) ),
	     this, SLOT( startDateChanged( const QString & ) ) );
    connect( picker, SIGNAL( dateClicked( int, int, int ) ),
	     this, SLOT( startDateChanged( int, int, int ) ) );

    QPopupMenu *m2 = new QPopupMenu( this );
    picker = new DateBookMonth( m2, 0, TRUE );
    m2->insertItem( picker );
    buttonEnd->setPopup( m2 );
    buttonEnd->setPopupDelay( 0 );
    endPicker = picker;
    connect( picker, SIGNAL( dateClicked( const QString & ) ),
	     this, SLOT( endDateChanged( const QString & ) ) );
    connect( picker, SIGNAL( dateClicked( int, int, int ) ),
	     this, SLOT( endDateChanged( int, int, int ) ) );

    comboCategory->clear();
    comboCategory->insertStringList( Global::categories() );
}

static QString fixDateString( const QString &s )
{
    QStringList l = QStringList::split( ' ', s );
    return l[1] + " " + l[2] + " " + l[3].right( 2 );
}

/*
 *  Destroys the object and frees any allocated resources
 */
DateEntry::~DateEntry()
{
    // no need to delete child widgets, Qt does it all for us
}

/*
 * public slot
 */
void DateEntry::endDateChanged( const QString &s )
{
    buttonEnd->setText( fixDateString( s ) );
}
/*
 * public slot
 */
void DateEntry::endDateChanged( int y, int m, int d )
{
    endDate.setYMD( y, m, d );
}
/*
 * public slot
 */
void DateEntry::endTimeChanged( const QString &s )
{
    QStringList l = QStringList::split( ':', s );
    endTime.setHMS( l[ 0 ].toInt(), l[ 1 ].toInt(), 0 );
}
/*
 * public slot
 */
void DateEntry::startDateChanged( const QString &s )
{
    buttonStart->setText( fixDateString( s ) );
}
/*
 * public slot
 */
void DateEntry::startDateChanged( int y, int m, int d )
{
    startDate.setYMD( y, m, d );
}
/*
 * public slot
 */
void DateEntry::startTimeChanged( const QString &s )
{
    QStringList l = QStringList::split( ':', s );
    startTime.setHMS( l[ 0 ].toInt(), l[ 1 ].toInt(), 0 );
}
/*
 * public slot
 */
void DateEntry::typeChanged( const QString &s )
{
    bool b = s != "All Day";
    buttonStart->setEnabled( b );
    buttonEnd->setEnabled( b );
    comboStart->setEnabled( b );
    comboEnd->setEnabled( b );
}


Event DateEntry::event()
{
    Event ev;
    ev.setDescription( comboDescription->currentText() );
    ev.setLocation( comboLocation->currentText() );
    ev.setCategory( comboCategory->currentText() );
    ev.setType( (Event::Type)comboType->currentItem() );
    if ( startDate > endDate ) {
	QDate tmp = endDate;
	endDate = startDate;
	startDate = tmp;
    }
    if ( startTime > endTime ) {
	QTime tmp = endTime;
	endTime = startTime;
	startTime = tmp;
    }
    ev.setStart( QDateTime( startDate, startTime ) );
    ev.setEnd( QDateTime( endDate, endTime ) );
    ev.setAlarm( checkAlarm->isChecked(), spinAlarm->value() );
    ev.setRepeat( checkRepeat->isChecked(), Event::RepeatPattern() ); // ### repeat pattern todo
    ev.setNotes( editNote->text() );
    return ev;
}
