/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include <qapplication.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qlabel.h>
#include "qimpenwidget.h"
#include "qimpensetup.h"
#include "qimpeninput.h"
#include "qimpenwordpick.h"
#include "qdawg.h"
#include "global.h"

#define ERROR_THRESHOLD	    140000
#define LOOKAHEAD_ERROR	    2500

// We'll use little pixmaps for the buttons to save screen space.

/* XPM */
static const char * const pen_xpm[] = {
"12 16 4 1",
" 	c None",
".	c #000000",
"+	c #FFFFFF",
"@	c #808080",
"            ",
"            ",
"         .  ",
"        .+. ",
"       ..@@.",
"      .+@.. ",
"     .+@@.  ",
"    .+@@.   ",
"   .+@@.    ",
"  .@.@.     ",
"  .@@.      ",
" ....       ",
" ..         ",
"            ",
"            ",
"            "};


/*!
  \class QIMPenInput qimpeninput.h

  Pen input widget.
*/
QIMPenInput::QIMPenInput( QWidget *parent, const char *name, WFlags f )
    : QFrame( parent, name, f )
{
    strokes.setAutoDelete( TRUE );
    charSets.setAutoDelete( TRUE );
    wordChars.setAutoDelete( TRUE );
    wordMatches.setAutoDelete( TRUE );

    QGridLayout *gl = new QGridLayout( this, 3, 2, 1, 1 );
    gl->setColStretch( 0, 1 );
    gl->setRowStretch( 2, 1 );

    wordPicker = new QIMPenWordPick( this );
    connect( wordPicker, SIGNAL(wordClicked(const QString &)),
	     this, SLOT(wordPicked(const QString &)) );
    QFont f("smallsmooth",9);
    QFontInfo fi( f );
    wordPicker->setFont( f );
    wordPicker->setBackgroundColor( white );
    gl->addRowSpacing( 0, fi.pointSize()+2 );
    gl->addMultiCellWidget( wordPicker, 0, 0, 0, 1 );

    pw = new QIMPenWidget( this );
    gl->addMultiCellWidget( pw, 1, 2, 0, 0 );

    QPixmap pm( (const char **)pen_xpm );
    setupBtn = new QPushButton( this );
    setupBtn->setFocusPolicy( NoFocus );
    setupBtn->setPixmap( pm );
    gl->addWidget( setupBtn, 1, 1 );
    connect( setupBtn, SIGNAL(clicked()), SLOT(setup()));

    connect( pw, SIGNAL(changeCharSet( QIMPenCharSet * )),
                 SLOT(selectCharSet( QIMPenCharSet * )) );
    connect( pw, SIGNAL(stroke( QIMPenStroke * )),
                 SLOT(strokeEntered( QIMPenStroke * )) );
    connect( pw, SIGNAL(beginStroke()), SLOT(beginStroke()) );

    matchChar = 0;
    currCharSet = 0;
    lastKey = 0;
    setupDlg = 0;
}

QIMPenInput::~QIMPenInput()
{
    wordMatches.clear();
    wordChars.clear();
}

QSize QIMPenInput::sizeHint() const
{
    int d = frameWidth();
    return pw->sizeHint() + QSize( 2*d, 2*d+9 );
}

/*!
  Add a character set from file \a fn.
*/
void QIMPenInput::addCharSet( const QString &fn )
{
    QIMPenCharSet *cs = new QIMPenCharSet( fn );
    if ( !cs->isEmpty() ) {
       charSets.append( cs );
       pw->addCharSet( cs );
       if ( !currCharSet )
	   currCharSet = cs;
    } else
       delete cs;
}

void QIMPenInput::wordPicked( const QString &w )
{
    int bs = wordChars.count();
    for ( int i = 0; i < bs; i++ ) {
	keypress( Qt::Key_Backspace << 16 );
    }

    for ( unsigned int i = 0; i < w.length(); i++ ) {
	keypress( w[i].unicode() );
    }

    wordMatches.clear();
    wordPicker->setWords( wordMatches );
}

void QIMPenInput::selectCharSet( QIMPenCharSet *cs )
{
    strokes.clear();
    testChar.clear();
    pw->clear();
    currCharSet = cs;
}

void QIMPenInput::beginStroke()
{
}

/*!
  Attempts to match the character entered and emit's key();
*/
void QIMPenInput::strokeEntered( QIMPenStroke *st )
{
    strokes.append( new QIMPenStroke( *st ) );
    testChar.addStroke( st );

    const QIMPenCharMatchList &ml = currCharSet->match( &testChar );
    if ( ml.getFirst() ) {
        matchChar = ml.getFirst()->penChar;
    }

    if ( !matchChar ) {
        //qDebug( "Unmatched strokes level 1 - panic" );
        strokes.clear();
        testChar.clear();
        pw->clear();
	return;
    }

    processMatches( ml );
}

void QIMPenInput::processMatches( const QIMPenCharMatchList &ml )
{
    if ( matchChar ) {
        if ( matchChar->penStrokes().count() == strokes.count() ) {
            if ( strokes.count() > 1 && lastKey
                 && (lastKey >> 16) != Qt::Key_Backspace) {
                //qDebug( "deleting last" );
                keypress( Qt::Key_Backspace << 16 );
		wordChars.removeLast();
		word.truncate( word.length() - 1 );
            }
	    updateWordMatch( ml );
            keypress( matchChar->character() );
            //qDebug( "** match \'%c\' ***", matchChar->character() );
            lastKey = matchChar->character();
        }

	bool greyed = FALSE;

	QListIterator<QIMPenCharMatch> it( ml );
	for ( ; it.current(); ++it ) {
	    QIMPenChar *m = it.current()->penChar;
	    if ( m->penStrokes().count() > matchChar->penStrokes().count() ) {
		greyed = TRUE;
		break;
	    }
	}

        if ( matchChar->penStrokes().count() < strokes.count() ) {
            for ( unsigned i = 0; i < matchChar->penStrokes().count(); i++ ) {
                strokes.removeFirst();
            }
            lastKey = 0;
            matchChar = 0;
            testChar.clear();
            QIMPenStroke *st = strokes.first();
            if ( st ) {
                testChar.addStroke( st );
                const QIMPenCharMatchList &ml2 = currCharSet->match( &testChar );
                if ( ml2.getFirst() ) {
                    matchChar = ml2.getFirst()->penChar;
                    if ( matchChar->penStrokes().count() == 1 ) {
			updateWordMatch( ml2 );
                        keypress( matchChar->character() );
                        //qDebug( "** match level 2 \'%c\' ***", matchChar->character() );
                        lastKey = matchChar->character();
                    }
                }

		QListIterator<QIMPenCharMatch> it2( ml2 );
		for ( ; it2.current(); ++it2 ) {
		    QIMPenChar *m = it2.current()->penChar;
		    if ( m->penStrokes().count() > matchChar->penStrokes().count() ) {
			greyed = TRUE;
			break;
		    }
		}
            }
        }
	pw->removeStroke( greyed );
    }
    if ( !matchChar && strokes.count() ) {
        //qDebug( "Unmatched strokes - panic" );
        strokes.clear();
        testChar.clear();
        pw->clear();
    }
}

void QIMPenInput::updateWordMatch( const QIMPenCharMatchList &ml )
{
    if ( !ml.count() )
	return;
    int ch = ml.getFirst()->penChar->character();
    QChar qch( ch );
    if ( qch.isPunct() || qch.isSpace() ||
	 (ch >> 16) == Qt::Key_Enter || (ch >> 16) == Qt::Key_Return ) {
	//qDebug( "Clearing word" );
	wordChars.clear();
	wordMatches.clear();
	word = QString();
    } else if ( ch >> 16 == Qt::Key_Backspace ) {
	//qDebug( "Handle backspace" );
	wordChars.removeLast();
	word.truncate( word.length() - 1 );
	matchWords();
    } else {
	QListIterator<QIMPenCharMatch> it( ml );
	QIMPenChar *matchCh;

	wordChars.append( new QIMPenCharMatchList() );
	wordChars.last()->setAutoDelete( TRUE );
	word += ml.getFirst()->penChar->character();

	for ( ; it.current(); ++it ) {
	    matchCh = it.current()->penChar;

	    if ( matchCh->penStrokes().count() == strokes.count() ) {
		QChar ch(matchCh->character());
		if ( !ch.isPunct() && !ch.isSpace() ) {
		    wordChars.last()->append( new QIMPenCharMatch( *it.current() ) );
		}
	    }
	}
	matchWords();
    }
    if ( !wordMatches.count() || wordMatches.getFirst()->word != word )
	wordMatches.prepend( new MatchWord( word, 0 ) );
    wordPicker->setWords( wordMatches );
}

void QIMPenInput::matchWords()
{
    if ( word.length() > 0 ) {
	// more leaniency if we don't have many matches
	if ( badMatches < 200 )
	    errorThreshold += (200 - badMatches) * 100;
    } else
	errorThreshold = ERROR_THRESHOLD;
    wordMatches.clear();
    goodMatches = 0;
    badMatches = 0;
    if ( wordChars.count() > 0 ) {
	maxGuess = (int)wordChars.count() * 2;
	if ( maxGuess < 3 )
	    maxGuess = 3;
	QString str;
	scanDict( Global::fixedDawg().root(), 0, str, 0 );
/*
	QListIterator<MatchWord> it( wordMatches);
	for ( ; it.current(); ++it ) {
	    qDebug( QString("Match word: %1").arg(it.current()->word) );
	}
*/
    }
    //qDebug( "Possibles: Good %d, total %d", goodMatches, wordMatches.count() );
    wordMatches.sort();
}

void QIMPenInput::scanDict( const QDawg::Node* n, int ipos, const QString& str, int error )
{
    if ( error / (ipos+1) > errorThreshold )
	return;

    while (n) {
	if ( goodMatches > 20 )
	    break;
	if ( ipos < (int)wordChars.count() ) {
	    QChar testCh = QChar(n->letter());
	    QListIterator<QIMPenCharMatch> it( *wordChars.at(ipos) );
	    for ( int i = 0; it.current() && i < 8; ++it, i++ ) {
		QChar ch( it.current()->penChar->character() );
		if ( ch == testCh || ( !ipos && ch.lower() == testCh.lower() ) ) {
		    int newerr =  error + it.current()->error;
		    if ( testCh.category() == QChar::Letter_Uppercase )
			ch = testCh;
		    QString newstr( str + ch );
		    if ( n->isWord() && ipos == (int)wordChars.count() - 1 ) {
			wordMatches.append( new MatchWord( newstr, newerr ) );
			goodMatches++;
		    }
		    scanDict( n->jump(), ipos+1, newstr, newerr );
		}
	    }
	} else if ( badMatches < 200 && ipos < maxGuess ) {
	    int d = ipos - wordChars.count();
	    int newerr = error + ERROR_THRESHOLD + LOOKAHEAD_ERROR*d;
	    QString newstr( str + n->letter() );
	    if ( n->isWord() ) {
		wordMatches.append( new MatchWord( newstr, newerr ) );
		badMatches++;
	    }
	    scanDict( n->jump(), ipos+1, newstr, newerr );
	}
	n = n->next();
    }
}

void QIMPenInput::keypress( unsigned int ch )
{
    int unicode = ch;

    if ( ch >= 'a' && ch <= 'z' ) {
	unicode |= ( Qt::Key_A + ch - 'a' ) << 16;
    } else if ( ch >= 'A' && ch <= 'Z' ) {
	unicode |= ( Qt::Key_A + ch - 'A' ) << 16;
    }

    emit key( unicode );
}

/*!
  Open the setup dialog
*/
void QIMPenInput::setup()
{
    if ( !setupDlg ) {
	// ### we are working with our copy of the char sets here.
	setupDlg = new QIMPenSetup( &charSets, 0, 0, TRUE );
	if ( qApp->desktop()->width() < 640 )
	    setupDlg->showMaximized();
	if ( setupDlg->exec() ) {
	    QListIterator<QIMPenCharSet> it(charSets);
	    for ( ; it.current(); ++it )
		it.current()->save();
	}
	delete setupDlg;
	setupDlg = 0;
    }
}


