/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "datebookmonth.h"
#include "pixmaploader.h"

#include <qtoolbutton.h>
#include <qspinbox.h>
#include <qcombobox.h>
#include <qdatetime.h>
#include <qpainter.h>

DateBookMonthHeader::DateBookMonthHeader( QWidget *parent, const char *name )
    : QHBox( parent, name )
{
    begin = new QToolButton( this );
    begin->setPixmap( PixmapLoader::loadPixmap( "start" ) );
    begin->setAutoRaise( TRUE );
    begin->setFixedSize( begin->sizeHint() );

    back = new QToolButton( this );
    back->setPixmap( PixmapLoader::loadPixmap( "back" ) );
    back->setAutoRaise( TRUE );
    back->setFixedSize( back->sizeHint() );

    month = new QComboBox( FALSE, this );
    for ( int i = 0; i < 12; ++i )
	month->insertItem( Calendar::nameOfMonth( i + 1 ) );

    year = new QSpinBox( 1900, 3000, 1, this );

    next = new QToolButton( this );
    next->setPixmap( PixmapLoader::loadPixmap( "forward" ) );
    next->setAutoRaise( TRUE );
    next->setFixedSize( next->sizeHint() );

    end = new QToolButton( this );
    end->setPixmap( PixmapLoader::loadPixmap( "finish" ) );
    end->setAutoRaise( TRUE );
    end->setFixedSize( end->sizeHint() );

    connect( month, SIGNAL( activated( int ) ),
	     this, SLOT( updateDate() ) );
    connect( year, SIGNAL( valueChanged( int ) ),
	     this, SLOT( updateDate() ) );
    connect( begin, SIGNAL( clicked() ),
	     this, SLOT( firstMonth() ) );
    connect( end, SIGNAL( clicked() ),
	     this, SLOT( lastMonth() ) );
    connect( back, SIGNAL( clicked() ),
	     this, SLOT( monthBack() ) );
    connect( next, SIGNAL( clicked() ),
	     this, SLOT( monthForward() ) );
    back->setAutoRepeat( TRUE );
    next->setAutoRepeat( TRUE );
}

void DateBookMonthHeader::updateDate()
{
    emit dateChanged( year->value(), month->currentItem() + 1 );
}

void DateBookMonthHeader::firstMonth()
{
    emit dateChanged( year->value(), 1 );
    month->setCurrentItem( 0 );
}

void DateBookMonthHeader::lastMonth()
{
    emit dateChanged( year->value(), 12 );
    month->setCurrentItem( 11 );
}

void DateBookMonthHeader::monthBack()
{
    if ( month->currentItem() > 0 ) {
	emit dateChanged( year->value(), month->currentItem() );
	month->setCurrentItem( month->currentItem() - 1 );
    } else {
	emit dateChanged( year->value() - 1, 12 );
	year->setValue( year->value() - 1 );
	month->setCurrentItem( 11 );
    }
}

void DateBookMonthHeader::monthForward()
{
    if ( month->currentItem() < 11 ) {
	emit dateChanged( year->value(), month->currentItem() + 2 );
	month->setCurrentItem( month->currentItem() + 1 );
    } else {
	emit dateChanged( year->value() + 1, 1 );
	year->setValue( year->value() + 1 );
	month->setCurrentItem( 0 );
    }
}

void DateBookMonthHeader::setDate( int y, int m )
{
    year->setValue( y );
    month->setCurrentItem( m - 1 );
}

DateBookMonthTable::DateBookMonthTable( QWidget *parent, const char *name )
    : QTable( 6, 7, parent, name )
{
    horizontalHeader()->setResizeEnabled( FALSE );
    for ( int i = 0; i < 7; ++i ) {
	horizontalHeader()->resizeSection( i, 30 );
	setColumnStretchable( i, TRUE );
	horizontalHeader()->setLabel( i, Calendar::nameOfDay( i + 1 ) );
    }

    verticalHeader()->hide();
    setLeftMargin( 0 );
    for ( int i = 0; i < 6; ++i )
	    setRowStretchable( i, TRUE );

    setSelectionMode( NoSelection );

    connect( this, SIGNAL( clicked( int, int, int, const QPoint & ) ),
	     this, SLOT( dayClicked( int, int ) ) );
    setVScrollBarMode( AlwaysOff );
    setHScrollBarMode( AlwaysOff );
}

void DateBookMonthTable::setupTable()
{
    QValueList<Calendar::Day> days = Calendar::daysOfMonth( year, month, FALSE );
    QValueList<Calendar::Day>::Iterator it = days.begin();
    int row = 0, col = 0;
    for ( ; it != days.end(); ++it ) {
	Calendar::Day day = *it;
	switch ( day.type ) {
	case Calendar::Day::PrevMonth:
	case Calendar::Day::NextMonth: {
	    DayItemMonth *i = new DayItemMonth( this, QTableItem::Never, "" );
	    i->setDay( day.date );
	    i->setType( day.type );
	    setItem( row, col, i );
	    if ( i->day() == this->day )
		setCurrentCell( row, col );
	    break;
	}
	case Calendar::Day::ThisMonth: {
	    DayItemMonth *i = new DayItemMonth( this, QTableItem::Never, "" );
	    i->setDay( day.date );
	    i->setType( day.type );
	    setItem( row, col, i );
	    if ( i->day() == this->day )
		setCurrentCell( row, col );
	    break;
	}	
	}

	if ( col == 6 ) {
	    ++row;
	    col = 0;
	} else {
	    ++col;
	}
    }
}

void DateBookMonthTable::dayClicked( int row, int col )
{
    DayItemMonth *i = (DayItemMonth*)item( row, col );
    if ( !i )
	return;
    selMonth = i->type() == Calendar::Day::ThisMonth ? month : i->type() == Calendar::Day::PrevMonth ? month - 1 : month + 1;

    selYear = year;
    if ( selMonth <= 0 ) {
	selMonth = 12;
	selYear--;
    } else if ( selMonth > 12 ) {
	selMonth = 1;
	selYear++;
    }
    selDay = i->day();
    emit dateClicked( selYear, selMonth,  selDay );
}

DateBookMonth::DateBookMonth( QWidget *parent, const char *name, bool ac )
    : QVBox( parent, name ), autoClose( ac )
{
    year = QDate::currentDate().year();
    month = QDate::currentDate().month();
    header = new DateBookMonthHeader( this, "DateBookMonthHeader" );
    table = new DateBookMonthTable( this, "DateBookMonthTable" );
    header->setDate( year, month );
    table->setDate( year, month, QDate::currentDate().day() );
    connect( header, SIGNAL( dateChanged( int, int ) ),
	     this, SLOT( setDate( int, int ) ) );
    connect( table, SIGNAL( dateClicked( int, int, int ) ),
	     this, SIGNAL( dateClicked( int, int, int ) ) );
    connect( table, SIGNAL( dateClicked( int, int, int ) ),
	     this, SLOT( forwardDateClicked( int, int, int ) ) );
    if ( ac )
	connect( table, SIGNAL( dateClicked( int, int, int ) ),
		 this, SLOT( setDate( int, int ) ) );
}

void DateBookMonth::setDate( int y, int m )
{
    year = y;
    month = m;
    table->setDate( year, month, 1 );
    if ( autoClose && parentWidget() && sender() == table )
	parentWidget()->close();
}

QDate  DateBookMonth::selectedDate() const
{
    if ( !table )
	return QDate::currentDate();
    int year,month,day;
    table->getDate( year, month, day );
    return QDate( year, month, day );
}


void DayItemMonth::paint( QPainter *p, const QColorGroup &cg, const QRect &cr, bool selected )
{
    QColorGroup g( cg );
    g.setBrush( QColorGroup::Base, back );
    g.setColor( QColorGroup::Text, forg );
    p->fillRect( 0, 0, cr.width(), cr.height(), selected ? g.brush( QColorGroup::Highlight ) : g.brush( QColorGroup::Base ) );

    if ( selected )
	p->setPen( g.highlightedText() );
    else
	p->setPen( g.text() );

    p->save();
    QFont f = p->font();
    f.setPointSize( ( f.pointSize() / 3 ) * 2 );
    p->setFont( f );
    QFontMetrics fm( f );
    p->drawText( 1, 1 + fm.ascent(), QString::number( day() ) );
    p->restore();
    // ##### draw some indication for the jobs of the day
}



void DayItemMonth::setType( Calendar::Day::Type t )
{
    switch ( t ) {
    case Calendar::Day::PrevMonth:
    case Calendar::Day::NextMonth:
	back = QBrush( darkBlue );
	forg = white;
	break;
    case Calendar::Day::ThisMonth:
	back = QBrush( white );
	forg = black;
	break;
    }
    typ = t;
}

