/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "minefield.h"
#include <qpainter.h>
#include <qpixmap.h>
#include <qtimer.h>
#include <stdlib.h>

static const char *pix_flag[]={
"12 12 3 1",
"# c #000000",
"x c #ff0000",
". c None",
"............",
"....#xxxxxx.",
"....#xxxxxx.",
"....#xxxxxx.",
"....#xxxxxx.",
"....#.......",
"....#.......",
"....#.......",
"....#.......",
"..#####.....",
".#######....",
"............"};

static const char *pix_mine[]={
"13 13 3 1",
"# c #000000",
". c None",
"a c #ffffff",
"......#......",
"......#......",
"..#.#####.#..",
"...#######...",
"..##aa#####..",
"..##aa#####..",
"#############",
"..#########..",
"..#########..",
"...#######...",
"..#.#####.#..",
"......#......",
"......#......"};

class Mine : public QTableItem
{
public:
    enum MineState {
	Hidden = 0,
	Empty,
	Mined,
	Flagged,
#ifdef MARK_UNSURE
	Unsure,
#endif
	Exploded,
	Wrong
    };

    Mine( QTable* );
    void paint( QPainter * p, const QColorGroup & cg, const QRect & cr, bool selected );
    EditType editType() const { return Never; }
    QSize sizeHint() const { return QSize( 14, 14 ); }

    void activate( bool sure = TRUE );
    void setHint( int );

    void setState( MineState );
    MineState state() const { return st; }

    bool isMined() const { return mined; }
    void setMined( bool m ) { mined = m; }
private:
    bool mined;
    int hint;

    MineState st;

    static QPixmap* knownField;
    static QPixmap* unknownField;
};

QPixmap* Mine::knownField = 0;
QPixmap* Mine::unknownField = 0;

Mine::Mine( QTable *t )
: QTableItem( t, Never, QString::null )
{
    mined = FALSE;
    st = Hidden;
    hint = 0;
}

void Mine::activate( bool sure )
{
    if ( !sure ) {
	switch ( st ) {
	case Hidden:
	    setState( Flagged );
	    break;
	case Flagged:
#ifdef MARK_UNSURE
	    setState( Unsure );
	    break;
	case Unsure:
#endif
	    setState( Hidden );
	default:
	    break;
	}
    } else if ( st == Flagged ) {
	return;
    } else {
	if ( mined ) {
	    setState( Exploded );
	} else {
	    setState( Empty );
	}
    }
}

void Mine::setState( MineState s )
{
    st = s;
}

void Mine::setHint( int h )
{
    hint = h;
}

void Mine::paint( QPainter* p, const QColorGroup &cg, const QRect& cr, bool )
{
    if ( !knownField ) {
	knownField = new QPixmap( cr.width(), cr.height() );
	QPainter pp( knownField );
	QBrush br( cg.button() );
	table()->style().drawButton( &pp, 0, 0, cr.width(), cr.height(), cg, TRUE, &br );
    }

    if ( !unknownField ) {
	unknownField = new QPixmap( cr.width(), cr.height() );
	QPainter pp( unknownField );
	QBrush br( cg.button() );
	table()->style().drawButton( &pp, 0, 0, cr.width(), cr.height(), cg, FALSE, &br );
    }

    p->save();

    switch(st) {
    case Hidden:
	p->drawPixmap( 0, 0, *unknownField );
	break;
    case Empty:
	p->drawPixmap( 0, 0, *knownField );
	if ( hint > 0 ) {
	    switch( hint ) {
	    case 1:
		p->setPen( blue );
		break;
	    case 2:
		p->setPen( green );
	    case 3:
		p->setPen( red );
		break;
	    default:
		p->setPen( darkMagenta );
		break;
	    }
	    p->drawText( QRect( 0, 0, cr.width(), cr.height() ), AlignHCenter | AlignVCenter, QString().setNum( hint ) );
	}
	break;
    case Mined:
	p->drawPixmap( 0, 0, *knownField );
	p->drawPixmap( 4, 4, QPixmap( pix_mine) );
	break;
    case Exploded:
	p->drawPixmap( 0, 0, *knownField );
	p->drawPixmap( 4, 4, QPixmap( pix_mine) );
	p->setPen( red );
	p->drawText( QRect( 0, 0, cr.width(), cr.height() ), AlignHCenter | AlignVCenter, "X" );
	break;
    case Flagged:
	p->drawPixmap( 0, 0, *unknownField );
	p->drawPixmap( 4, 4, QPixmap( pix_flag ) );
	break;
#ifdef MARK_UNSURE
    case Unsure:
	p->drawPixmap( 0, 0, *unknownField );
	p->drawText( QRect( 0, 0, cr.width(), cr.height() ), AlignHCenter | AlignVCenter, "?" );
	break;
#endif
    case Wrong:
	p->drawPixmap( 0, 0, *unknownField );
	p->drawPixmap( 4, 4, QPixmap( pix_flag) );
	p->setPen( red );
	p->drawText( QRect( 0, 0, cr.width(), cr.height() ), AlignHCenter | AlignVCenter, "X" );
	break;
    }

    p->restore();
}

/*
  MineField implementation
*/

MineField::MineField( QWidget* parent, const char* name )
: QTable( parent, name )
{
    setState( GameOver );
    setShowGrid( FALSE );
    horizontalHeader()->hide();
    verticalHeader()->hide();
    setTopMargin( 0 );
    setLeftMargin( 0 );

    setSizePolicy( QSizePolicy( QSizePolicy::Maximum, QSizePolicy::Maximum ) );
    
    setSelectionMode( QTable::NoSelection );
    setFocusPolicy( QWidget::NoFocus );

    setCurrentCell( -1, -1 );

    connect( this, SIGNAL( pressed( int, int, int, const QPoint& ) ), this, SLOT( cellPressed( int, int ) ) );
    connect( this, SIGNAL( clicked( int, int, int, const QPoint& ) ), this, SLOT( cellClicked( int, int ) ) );

    holdTimer = new QTimer( this );
    connect( holdTimer, SIGNAL( timeout() ), this, SLOT( held() ) );

    flagAction = NoAction;
    ignoreClick = FALSE;
    currRow = currCol = 0;
}

MineField::~MineField()
{
}

void MineField::setState( State st )
{
    stat = st;
}


void MineField::setup( int level )
{
    setState( Waiting );
    viewport()->setUpdatesEnabled( FALSE );

    int x; 
    int y;
    for ( x = 0; x < numCols(); x++ )
	for ( y = 0; y < numRows(); y++ )
	    clearCell( y, x );

    switch( level ) {
    case 1:
	setNumRows( 9 );
	setNumCols( 9 );
	minecount = 12;
	break;
    case 2:
	setNumRows( 16 );
	setNumCols( 16 );
	minecount = 40;
	break;
    case 3:
	setNumRows( 30 );
	setNumCols( 16 );
	minecount = 99;
	break;
    }
    nonminecount = numRows()*numCols() - minecount;
    mineguess = minecount;
    emit mineCount( mineguess );
    
    for ( x = 0; x < numCols(); x++ ) {
	for ( y = 0; y < numRows(); y++ ) {
	    setItem( y, x, new Mine( this ) );
	    if ( !x )
		setRowHeight( y, 20 );
	}
	setColumnWidth( x, 20 );
    }

    updateGeometry();
    viewport()->setUpdatesEnabled( TRUE );
    viewport()->repaint( TRUE );
}


void MineField::placeMines()
{
    int mines = minecount;
    while ( mines ) {
	int col = int((double(rand()) / double(RAND_MAX)) * numCols());
	int row = int((double(rand()) / double(RAND_MAX)) * numRows());

	Mine* mine = (Mine*)item( row, col );

	if ( mine && !mine->isMined() && mine->state() == Mine::Hidden ) {
	    mine->setMined( TRUE );
	    mines--;
	}
    }
}

void MineField::paintFocus( QPainter*, const QRect& )
{
}

void MineField::viewportMousePressEvent( QMouseEvent* e )
{
    QTable::viewportMousePressEvent( e );
}

void MineField::viewportMouseReleaseEvent( QMouseEvent* e )
{
    QTable::viewportMouseReleaseEvent( e );
    if ( flagAction == FlagNext ) {
	flagAction = NoAction;
    }
}

void MineField::keyPressEvent( QKeyEvent* e )
{
#if defined(_WS_QWS_)
    flagAction = ( e->key() == Key_Up ) ? FlagOn : NoAction;
#else
    flagAction = ( ( e->state() & ShiftButton ) ==  ShiftButton ) ? FlagOn : NoAction;
#endif
}

void MineField::keyReleaseEvent( QKeyEvent* )
{
    flagAction = NoAction;
}

int MineField::getHint( int row, int col )
{
    int hint = 0;
    for ( int c = col-1; c <= col+1; c++ )
	for ( int r = row-1; r <= row+1; r++ ) {
	    Mine* mine = (Mine*)item( r, c );
	    if ( mine && mine->isMined() )
		hint++;
	}

    return hint;
}

void MineField::setHint( Mine* mine )
{
    if ( !mine )
	return;

    int row = mine->row();
    int col = mine->col();
    int hint = getHint( row, col );

    if ( !hint ) {
	for ( int c = col-1; c <= col+1; c++ )
	    for ( int r = row-1; r <= row+1; r++ ) {
		Mine* mine = (Mine*)item( r, c );
		if ( mine && mine->state() == Mine::Hidden ) {
		    mine->activate( TRUE );
		    nonminecount--;
		    setHint( mine );
		    updateCell( r, c );
		}
	    }
    }

    mine->setHint( hint );
    updateCell( row, col );
}

/*
 state == Waiting means no "hold"

 
*/
void MineField::cellPressed( int row, int col )
{
    if ( state() == GameOver ) 
	return;
    currRow = row;
    currCol = col;
    if ( state() == Playing )
	holdTimer->start( 250, TRUE );
}

void MineField::held()
{
    flagAction = FlagNext;
    updateMine( currRow, currCol );
    ignoreClick = TRUE;
}

/*
  Only place mines after first click, since it is pointless to
  kill the player before the game has started.
*/

void MineField::cellClicked( int row, int col )
{
    if ( state() == GameOver )
	return;
    if ( state() == Waiting ) {
	Mine* mine = (Mine*)item( row, col );
	if ( !mine )
	    return;
	mine->setState( Mine::Empty );
	nonminecount--;
	placeMines();
	setState( Playing );
	emit gameStarted();
	updateMine( row, col );
    } else { // state() == Playing
	holdTimer->stop();
	if ( ignoreClick )
	    ignoreClick = FALSE;
	else
	    updateMine( row, col );
    }
}

void MineField::updateMine( int row, int col )
{
    Mine* mine = (Mine*)item( row, col );
    if ( !mine )
	return;

    bool wasFlagged = mine->state() == Mine::Flagged;
    bool wasEmpty =  mine->state() == Mine::Empty;
    
    mine->activate( flagAction == NoAction );

    if ( mine->state() == Mine::Exploded ) {
	emit gameOver( FALSE );
	setState( GameOver );
	return;
    } else if ( mine->state() == Mine::Empty ) {
	setHint( mine );
	if ( !wasEmpty )
	    nonminecount--;
    }

    if ( flagAction != NoAction ) {
	if ( mine->state() == Mine::Flagged ) {
	    --mineguess;
	    emit mineCount( mineguess );
	    if ( mine->isMined() )
		--minecount;
	} else if ( wasFlagged ) {
	    ++mineguess;
	    emit mineCount( mineguess );
	    if ( mine->isMined() )
		++minecount;
	}
    }

    updateCell( row, col );

    if ( !minecount && !mineguess || !nonminecount ) {
	emit gameOver( TRUE );
	setState( GameOver );
    }
}

void MineField::showMines()
{
    for ( int c = 0; c < numCols(); c++ )
	for ( int r = 0; r < numRows(); r++ ) {
	    Mine* mine = (Mine*)item( r, c );
	    if ( !mine )
		continue;
	    if ( mine->isMined() && mine->state() == Mine::Hidden )
		mine->setState( Mine::Mined );
	    if ( !mine->isMined() && mine->state() == Mine::Flagged )
		mine->setState( Mine::Wrong );

	    updateCell( r, c );
	}
}
