#include <qapplication.h>
#include <qmessagebox.h>
#include <qcombobox.h>
#include <qdatetime.h>
#include <qfileinfo.h>
#include <qiconset.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qtextstream.h>
#include <qtimer.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qvbox.h>
#include <qwidgetstack.h>
#include <qpainter.h>
#include <qregexp.h>

#include "wordgame.h"
#include "global.h"
#include "filemanager.h"
#include "pixmaploader.h"

#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>


enum RuleEffects {
    Multiplier=15,
    MultiplyAll=64,
    Start=128
};

const char* sampleWGR=
	"wordgame_shapes\n"
	"15 15\n"
	"400001040100004\n"
	"030000000000030\n"
	"002002000200200\n"
	"000300020003000\n"
	"000020000020000\n"
	"102001000100201\n"
	"000000202000000\n"
	"400200050002004\n"
	"000000202000000\n"
	"102001000100201\n"
	"000020000020000\n"
	"000300020003000\n"
	"002002000200200\n"
	"030000000000030\n"
	"400001040100004\n"
	"1 2 3 66 67 194 0\n"
	"1 j 8\n"
	"1 q 7\n"
	"1 x 6\n"
	"1 z 6\n"
	"1 w 4\n"
	"1 k 4\n"
	"1 v 3\n"
	"1 f 3\n"
	"2 y 3\n"
	"2 h 2\n"
	"2 b 2\n"
	"2 m 2\n"
	"3 p 2\n"
	"3 g 2\n"
	"3 u 2\n"
	"4 d 2\n"
	"4 c 2\n"
	"5 l 1\n"
	"5 o 1\n"
	"7 t 1\n"
	"7 n 1\n"
	"7 a 1\n"
	"7 r 1\n"
	"8 s 1\n"
	"8 i 1\n"
	"11 e 1\n"
	"0\n";

WordGame::WordGame( QWidget* parent, const char* name, WFlags fl ) :
    QMainWindow(parent, name, fl)
{
    setCaption( "Word Game" );
    setToolBarsMovable( FALSE );
    vbox = new QVBox(this);

    setCentralWidget(vbox);
    QToolBar* toolbar = new QToolBar(this);
    addToolBar(toolbar, Bottom);
    new QToolButton(PixmapLoader::loadPixmap("close"), "Close", "", this, SLOT(endGame()), toolbar);
    done = new QToolButton(PixmapLoader::loadPixmap("done"), "Done", "", this, SLOT(endTurn()), toolbar);
    reset = new QToolButton(PixmapLoader::loadPixmap("reset"), "Reset", "", this, SLOT(resetTurn()), toolbar);
    pass = new QToolButton(PixmapLoader::loadPixmap("pass"), "Pass", "", this, SLOT(passTurn()), toolbar);
    scoreinfo = new ScoreInfo(toolbar);
    scoreinfo->setFont(QFont("Helvetica",10));

    cpu = 0;
    board = 0;
    bag = 0;
    racks = 0;

    aiheart = new QTimer(this);
    connect(aiheart, SIGNAL(timeout()), this, SLOT(think()));

    startGame();
}

void WordGame::closeEvent(QCloseEvent* e)
{
    if ( gameover || mayEndGame() ) {
	e->accept();
	return;
    }
}

void WordGame::startGame()
{
    gameover = FALSE;

    delete board;
    board = 0;
    delete racks;
    racks = 0;

    delete cpu;
    cpu = 0;
    nplayers=0;

    NewGame game(this);
    //Rules rules(this);
    //connect(game.editrules, SIGNAL(clicked()), &rules, SLOT(editRules()));
    //connect(&rules, SIGNAL(rulesChanged()), &game, SLOT(updateRuleSets()));
    struct passwd* n = getpwuid(getuid());
    QString playername = n ? n->pw_name : "";
    if ( playername.isEmpty() ) {
	playername = "Player";
    }
    game.player0->changeItem(playername,0);
    game.player1->setCurrentItem(1);
    game.updateRuleSets();
#ifndef _WS_QWS
    game.showMaximized();
#endif
    if ( game.exec() ) {
	if ( loadRules(game.rules->currentText()) ) {
	    racks = new QWidgetStack(vbox);
	    namelist.clear();
	    addPlayer(game.player0->currentText());
	    addPlayer(game.player1->currentText());
	    addPlayer(game.player2->currentText());
	    addPlayer(game.player3->currentText());
	    addPlayer(game.player4->currentText());
	    addPlayer(game.player5->currentText());
	    scoreinfo->init(namelist);

	    if ( nplayers ) {
		player=0;
		readyRack(player);
	    }

	    board->show();
	    racks->show();
	}
    }
}

bool WordGame::loadRules(const QString& fn)
{
    setCaption( fn );
    QByteArray ba; // #### encapsulate this bit in FileManager
    FileManager fm;
    FileManager::File f;
    f.filename = fn + ".WGR"; // wordgame rules
    f.category = "Private";
    if ( !fm.loadFile(f,ba) )
	return FALSE;
    QTextStream ts(ba,IO_ReadOnly);


    QString shapepixmap;
    ts >> shapepixmap;
    int htiles,vtiles;
    ts >> htiles >> vtiles;

    if ( htiles < 3 || vtiles < 3 )
	return FALSE;

    QPixmap bgshapes = PixmapLoader::loadPixmap(shapepixmap);
    QString rule_shapes;
    for (int i=0; i<vtiles; i++) {
	QString line;
	ts >> line;
	rule_shapes += line;
    }
    static int rule_effects[10];
    int re=0,e;
    ts >> e;
    while ( e && re < 10 ) {
	rule_effects[re] = e;
	if ( re++ < 10 ) ts >> e;
    }
    board = new Board(bgshapes, htiles, vtiles, vbox);
    board->setRules(rule_shapes, rule_effects);
    connect(board, SIGNAL(temporaryScore(int)), scoreinfo, SLOT(showTemporaryScore(int)));

    bag = new Bag;

    int count;
    ts >> count;
    while ( count ) {
	QString text;
	int value;
	ts >> text >> value;

	Tile t(text, value);
	for (int n=count; n--; )
	    bag->add(t);

	ts >> count;
    }

    return TRUE;
}


NewGame::NewGame(QWidget* parent) :
    NewGameBase(parent,0,TRUE)
{
}

void NewGame::updateRuleSets()
{
    rules->clear();

    FileManager fm;
    QValueList<FileManager::File> files = fm.files("Private", "*.WGR");
    if ( files.count() == 0 ) {
	// Provide a sample
	FileManager::File f;
	f.filename = "Sample.WGR";
	f.category = "Private";
	QByteArray ba(strlen(sampleWGR));
	memcpy(ba.data(),sampleWGR,strlen(sampleWGR));
	fm.saveFile(f,ba);
	files += f;
    }
    int newest=0;
    int newest_age=INT_MAX;
    QDateTime now = QDateTime::currentDateTime();
    for (QValueList<FileManager::File>::Iterator it=files.begin(); it!=files.end(); ++it) {
	QFileInfo fi(fm.absoluteFileName(*it));
	int age = fi.lastModified().secsTo(now);
	QString name = (*it).filename;
	name.replace(QRegExp(".WGR$"),"");
	rules->insertItem(fi.baseName());
	if ( age < newest_age ) {
	    newest_age = age;
	    newest = rules->count()-1;
	}
    }
    rules->setCurrentItem(newest);
}

Rules::Rules(QWidget* parent) :
    RulesBase(parent,0,TRUE)
{
}

void Rules::editRules()
{
    if ( exec() ) {
	// ### create a new set of rules
	emit rulesChanged();
    }
}

void Rules::deleteRuleSet()
{
    // ### delete existing rule set
    emit rulesChanged();
}

void WordGame::addPlayer(const QString& name)
{
    if ( !name.isEmpty() ) {
	int colon = name.find(':');
	int cpu = (colon >=0 && name.left(2) == "AI") ? name.mid(2,1).toInt() : 0;
	addPlayer(name,cpu);
    }
}

void WordGame::addPlayer(const QString& name, int cpu)
{
    static const int rack_tiles=7;

    Rack* r = new Rack(rack_tiles,racks);
    r->setPlayerName(name);
    r->setComputerization(cpu);
    racks->addWidget(r, nplayers);
    refillRack(nplayers);
    namelist.append(name);

    ++nplayers;
}

void WordGame::nextPlayer()
{
    if ( !refillRack(player) ) {
	endGame();
    } else {
	player = (player+1)%nplayers;
	scoreinfo->setBoldOne(player);
	readyRack(player);
    }
}

bool WordGame::mayEndGame()
{
    int out=-1;
    int i;
    for (i=0; i<nplayers; i++)
	if ( !rack(i)->count() )
	    out = i;
    if ( out<0 ) {
	if ( QMessageBox::warning(this,"End game",
	    "Do you want to end the game early?", "Yes", "No" )!=0 )
	{
	    return FALSE;
	}
    }
    return TRUE;
}

void WordGame::endGame()
{
    if ( gameover ) {
	close();
	return;
    }

    if ( !mayEndGame() )
	return;
    int out=-1;
    int totalleft=0;
    int i;
    for (i=0; i<nplayers; i++) {
	Rack* r = rack(i);
	int c = r->count();
	if ( c ) {
	    int lose=0;
	    for ( int j=0; j<c; j++ )
		lose += r->tileRef(j)->value();
	    totalleft += lose;
	    scoreinfo->addScore(i,-lose);
	} else {
	    out = i;
	}
    }
    int highest=0;
    int winner=0;
    for (i=0; i<nplayers; i++) {
	int s = scoreinfo->playerScore(i);
	if ( s > highest ) {
	    highest = s;
	    winner = i;
	}
    }
    if ( out >= 0 )
	scoreinfo->addScore(out,totalleft);
    scoreinfo->setBoldOne(winner);
    gameover = TRUE;
    done->setEnabled(TRUE);
    reset->setEnabled(FALSE);
    pass->setEnabled(FALSE);
}

void WordGame::endTurn()
{
    if ( gameover ) {
	startGame();
    } else {
	if ( board->turnScore() >= 0 ) {
	    scoreinfo->addScore(player,board->turnScore());
	    board->placeFloaters();
	    nextPlayer();
	} else {
	    QApplication::beep();
	}
    }
}

void WordGame::resetTurn()
{
    board->resetRack();
}

void WordGame::passTurn()
{
    // ######## trade?
    nextPlayer();
}

bool WordGame::refillRack(int i)
{
    Rack* r = rack(i);
    while ( !bag->isEmpty() && !r->isFull() ) {
	r->addTile(bag->takeRandom());
    }
    return r->count() != 0;
}

void WordGame::readyRack(int i)
{
    Rack* r = rack(i);
    racks->raiseWidget(i);
    board->setCurrentRack(r);

    done->setEnabled( !r->computerized() );
    reset->setEnabled( !r->computerized() );
    pass->setEnabled( !r->computerized() );

    if ( r->computerized() ) {
	cpu = new ComputerPlayer(board, r);
	aiheart->start(0);
    }
}

Rack* WordGame::rack(int i) const
{
    return (Rack*)racks->widget(i);
}

void WordGame::think()
{
    if ( !cpu->step() ) {
	delete cpu;
	cpu = 0;
	aiheart->stop();
	if ( board->turnScore() < 0 )
	    passTurn();
	else
	    endTurn();
    }
}

bool ComputerPlayer::step()
{
    const QDawg::Node* root = Global::fixedDawg().root();
    QPoint d = across ? QPoint(1,0) : QPoint(0,1);
    signed char rev[4096]; // QDawg only supports 0..4095 chars
    memset(rev,-1,4096);
    for (int i=0; i<rack->count(); i++) {
	uint x = rack->tile(i).text()[0].unicode();
	ASSERT (x<4096);
	rev[x] = i;
    }
    const Tile* tiles[99]; // ### max board size
    findBest(current, d, root, 0, rev, tiles, 0);
    if ( ++current.rx() == board->xTiles() ) {
	current.rx() = 0;
	if ( ++current.ry() == board->yTiles() ) {
	    if ( across ) {
		if ( best_score >= 0 ) {
		    rack->arrangeTiles(best);
		    board->scoreTurn(best_start, best.length(), best_dir);
		    board->placeTiles(best_start, best.length(), best_dir);
		}
		return FALSE;
	    } else {
		across = TRUE;
		current = QPoint(0,0);
	    }
	}
    }
    return TRUE;
}

void ComputerPlayer::findBest(QPoint at, const QPoint& d, const QDawg::Node* node, ulong used, const signed char *rev, const Tile** tiles, int n)
{
    if ( !node ) {
	return;
    }
    QChar l = node->letter();
    const Tile* cur = board->tile(at);
    if ( cur ) {
	if ( cur->text()[0] == l ) {
	    bool nextok =  board->contains(at+d);
	    if ( node->isWord() && n && (!nextok || !board->tile(at+d)) )
		noteChoice(tiles,n,d);
	    if ( nextok )
		findBest(at+d, d, node->jump(), used, rev, tiles, n);
	    // ###### text()[1]...
	}
    } else {
	int x = rev[l.unicode()];
	if ( x >= 0 && !(used&(1<<x))) {
	    bool nextok =  board->contains(at+d);
	    tiles[n++] = rack->tileRef(x);
	    if ( node->isWord() )
	    if ( node->isWord() && (!nextok || !board->tile(at+d)) )
		noteChoice(tiles,n,d);
	    used |= 1<<x; // mark
	    if ( nextok )
		findBest(at+d, d, node->jump(), used, rev, tiles, n);
	    n--;
	}
	// ###### text()[1]...
    }
    findBest(at, d, node->next(), used, rev, tiles, n);
}

void ComputerPlayer::noteChoice(const Tile** tiles, int n, const QPoint& d)
{
    int s = board->score(current, tiles, n, d, 0);
    if ( s > best_score ) {
	best = "";
	for ( int i=0; i<n; i++ )
	    best += tiles[i]->text();
	best_score = s;
	best_dir = d;
	best_start = current;
    }
}

int TileItem::smallWidth()
{
    return 16;
}

int TileItem::smallHeight()
{
    return 16;
}

int TileItem::bigWidth()
{
    return 22;
}

int TileItem::bigHeight()
{
    return 22;
}

void TileItem::setState( State state )
{
    hide();
    s = state;
    show(); // ### use update() in Qt 3.0
}

void TileItem::setTile(const Tile& tile)
{
    hide();
    t = tile;
    show(); // ### use update() in Qt 3.0
}

void TileItem::setBig(bool b)
{
    big = b;
}

void TileItem::drawShape(QPainter& p)
{
    static QFont value_font("heletica",8);
    static QFont big_font("smoothtimes",17);
    static QFont small_font("smoothtimes",10);

    QRect area(x(),y(),width(),height());
    p.setBrush(s == Floating ? lightGray : white);
    p.drawRect(area);
    if ( big ) {
	p.setFont(value_font);
	QString n = QString::number(t.value());
	int w = p.fontMetrics().width('1');
	int h = p.fontMetrics().height();
	w *= n.length();
	QRect valuearea(x()+width()-w-2,y()+height()-h+1,w,h);
	p.drawText(valuearea,AlignCenter,n);
	p.setFont(big_font);
	area = QRect(x(),y(),width()-2,height()-1);
    } else {
	p.setFont(small_font);
	area = QRect(x(),y()+2,width(),height()-2);
    }
    p.drawText(area,AlignCenter,t.text().upper());
}

Board::Board(QPixmap bgshapes, int w, int h, QWidget* parent) :
    QCanvasView(new QCanvas(bgshapes,w,h, TileItem::smallWidth(), TileItem::smallHeight()),
		parent)
{
    grid = new TileItem*[w*h];
    memset(grid,0,w*h*sizeof(TileItem*));
    setFrameStyle(0);
    setHScrollBarMode(AlwaysOff);
    setVScrollBarMode(AlwaysOff);
    current_rack = 0;
}

Board::~Board()
{
    delete canvas();
}

void Board::setCurrentRack(Rack* r)
{
    turn_score = -1;
    current_rack = r;
}

void Board::resetRack()
{
    removeFloaters();
    canvas()->update();
}

void Board::contentsMousePressEvent(QMouseEvent* e)
{
    dragstart = e->pos();
}

void Board::contentsMouseMoveEvent(QMouseEvent* e)
{
    if ( current_rack && !current_rack->computerized() ) {
	QPoint d = e->pos() - dragstart;
	if ( d.x() <= 0 && d.y() <= 0 ) {
	    // None
	    resetRack();
	} else {
	    int n;
	    QPoint start=boardPos(dragstart);
	    QPoint end=boardPos(e->pos());
	    QPoint diff=end-start;
	    QPoint dir;
	    if ( d.x() > d.y() ) {
		n = diff.x()+1;
		dir = QPoint(1,0);
	    } else {
		n = diff.y()+1;
		dir = QPoint(0,1);
	    }

	    removeFloaters();

	    // Subtract existing tiles from n
	    QPoint t = start;
	    for ( int i=n; i--; ) {
		if ( contains(t) && tile(t) )
		    n--;
		t += dir;
	    }

	    // Move start back to real start
	    while (contains(start-dir) && tile(start-dir))
		start -= dir;

	    scoreTurn(start, n, dir);
	    placeTiles(boardPos(dragstart), n, dir);
	}
    }
}

void Board::placeFloaters()
{
    // ## if we remembered where they were, this could be faster
    int n = xTiles()*yTiles();
    for (int i=0; i<n; i++)
	if ( grid[i] && grid[i]->state()==TileItem::Floating ) {
	    current_rack->remove(grid[i]->tile());
	    grid[i]->setState(TileItem::Firm);
	}
    canvas()->update();
}

void Board::removeFloaters()
{
    // ## if we remembered where they were, this could be faster
    int n = xTiles()*yTiles();
    for (int i=0; i<n; i++)
	if ( grid[i] && grid[i]->state()==TileItem::Floating ) {
	    delete grid[i];
	    grid[i] = 0;
	}
}

void Board::placeTiles(QPoint at, int count, const QPoint& step)
{
    int i=0;
    removeFloaters();
    while ( i<count && i<current_rack->count() && contains(at) ) {
	if ( !tile(at) ) {
	    Tile t = current_rack->tile(i);
	    setTile(at,t);
	    setTileState(at,TileItem::Floating);
	    i++;
	}
	at += step;
    }
    canvas()->update();
}

int Board::bonussedValue(const QPoint& at, int base, int& all_mult) const
{
    int rule = rule_shape[idx(at)]-'0';
    int effect = rule_effect[rule];
    int mult = effect&Multiplier;
    if ( effect & MultiplyAll ) {
	all_mult *= mult;
	return base;
    } else {
	return base * mult;
    }
}

bool Board::isStart(const QPoint& at) const
{
    int rule = rule_shape[idx(at)]-'0';
    int effect = rule_effect[rule];
    return effect&Start;
}

void Board::scoreTurn(const QPoint& at, int n, const QPoint& d)
{
    removeFloaters();
    const Tile* tiles[99];
    for (int i=0; i<n; i++)
	tiles[i] = current_rack->tileRef(i);
    turn_score = score(at,tiles,n,d,&turn_word);
    emit temporaryScore(turn_score);
}

int Board::score(QPoint at, const Tile** tiles, int n, const QPoint& d, QString* mainword) const
{
    int total=0;
    int totalsidetotal=0;

    // mainword==0 ->
    // Checks side words, but not main word

    // -1 means bad words

    // text is assumed to fit on board.

    QPoint otherd(d.y(), d.x());

    int all_mult = 1;

    bool connected = FALSE;

    if ( mainword )
	*mainword="";

    if ( contains(at-d) && tile(at-d) ) {
	return -1; // preceeding tiles
    }

    const Tile* t;
    for (int i=0; contains(at) && ((t=tile(at)) || i<n); ) {
	if ( t ) {
	    if ( mainword )
		*mainword += t->text();
	    total += t->value();
	    connected = TRUE;
	} else {
	    QString word = tiles[i]->text();
	    if ( mainword )
		*mainword += word;
	    int side_mult = 1;
	    int tilevalue = bonussedValue(at,tiles[i]->value(),side_mult);
	    all_mult *= side_mult;
	    if ( !connected && isStart(at) )
		connected = TRUE;
	    total += tilevalue;
	    int sidetotal = tilevalue;
	    {
		QPoint side = at-otherd;
		while ( contains(side) && (t=tile(side)) ) {
		    sidetotal += t->value();
		    word.prepend(t->text());
		    side -= otherd;
		}
	    }
	    {
		QPoint side = at+otherd;
		while ( contains(side) && (t=tile(side)) ) {
		    sidetotal += t->value();
		    word.append(t->text());
		    side += otherd;
		}
	    }
	    if ( word.length()>1 ) {
		if ( !Global::fixedDawg().contains(word) )
		    return -1;
		totalsidetotal += sidetotal * side_mult;
		connected = TRUE;
	    }
	    i++;
	}
	at += d;
    }

    if ( mainword && !Global::fixedDawg().contains(*mainword) )
	return -1;

    return connected ? totalsidetotal + total * all_mult : -1;
}

QPoint Board::boardPos(const QPoint& p) const
{
    return QPoint(p.x()/canvas()->tileWidth(), p.y()/canvas()->tileHeight());
}

void Board::contentsMouseReleaseEvent(QMouseEvent*)
{
    if ( current_rack ) {
    }
}


void Board::setRules(const QString& shapes, const int* effects)
{
    rule_shape=shapes; rule_effect=effects;
    int i=0;
    for (int y=0; y<yTiles(); y++)
	for (int x=0; x<xTiles(); x++)
	    canvas()->setTile(x,y,shapes[i++]-'0');
}

void Board::unsetTile(const QPoint& p)
{
    delete item(p);
    grid[idx(p)] = 0;
}

void Board::setTile(const QPoint& p, const Tile& t)
{
    TileItem* it=item(p);
    if ( !it ) {
	it = grid[idx(p)] = new TileItem(t,FALSE,canvas());
	it->move(p.x()*canvas()->tileWidth(), p.y()*canvas()->tileHeight());
    } else {
	it->setTile(t);
    }
}

Rack::Rack(int ntiles, QWidget* parent) : QCanvasView(
	new QCanvas(ntiles*TileItem::bigWidth(),TileItem::bigHeight()),
	    parent),
	item(ntiles)
{
    setLineWidth(1);
    setFixedHeight(sizeHint().height());
    n = 0;
    for (int i=0; i<ntiles; i++)
	item[i]=0;
    setHScrollBarMode(AlwaysOff);
    setVScrollBarMode(AlwaysOff);
    canvas()->setBackgroundColor(gray);
    dragging = 0;
}

Rack::~Rack()
{
    for (int i=0; i<n; i++)
	delete item[i];
    delete canvas();
}

static int cmp_tileitem(const void *a, const void *b)
{
    const TileItem* ia = *(TileItem**)a;
    const TileItem* ib = *(TileItem**)b;
    return int(ia->x() - ib->x());
}

void Rack::layoutTiles()
{
    int w = TileItem::bigWidth()+2;

    if ( dragging ) dragging->moveBy(dragging_adj,0);
    qsort(item.data(), n, sizeof(TileItem*), cmp_tileitem);
    if ( dragging ) dragging->moveBy(-dragging_adj,0);

    for (int i=0; i<n ;i++)
	if ( item[i] == dragging ) {
	    item[i]->setZ(1);
	} else {
	    item[i]->move(i*w, 0);
	    item[i]->setZ(0);
	}
    canvas()->update();
}

bool Rack::arrangeTiles(QString s)
{
    bool could = TRUE;
    for (int j=0; j<n; j++) {
	QString tt = item[j]->tile().text();
	int f = s.find(tt);
	if ( f >= 0 ) {
	    s.replace(f,tt.length(),"#");
	    item[j]->move(f-999,0);
	} else {
	    could = FALSE;
	}
    }
    layoutTiles();
    return could;
}

void Rack::addTile(const Tile& t)
{
    TileItem *i = new TileItem(t,TRUE,canvas());
    i->show();
    item[n++] = i;
    layoutTiles();
}

void Rack::remove(Tile t)
{
    for (int i=0; i<n ;i++)
	if ( item[i]->tile() == t ) {
	    remove(i);
	    return;
	}
}

void Rack::remove(int i)
{
    delete item[i];
    n--;
    for (;i<n;i++)
	item[i]=item[i+1];
    layoutTiles();
}

void Rack::resizeEvent(QResizeEvent* e)
{
    canvas()->resize(width()-frameWidth()*2,height()-frameWidth()*2);
    QCanvasView::resizeEvent(e);
}

void Rack::contentsMousePressEvent(QMouseEvent* e)
{
    if ( computerized() )
	return;
    QCanvasItemList list = canvas()->collisions(e->pos());
    if (list.count()) {
	dragging = list.first();
	dragstart = e->pos()-QPoint(int(dragging->x()),int(dragging->y()));
    } else {
	dragging = 0;
    }
}

void Rack::contentsMouseMoveEvent(QMouseEvent* e)
{
    if ( computerized() )
	return;
    int w = TileItem::bigWidth()+2;
    if ( dragging ) {
	dragging_adj = TileItem::bigWidth()/2;
	if ( dragging->x() > e->x()-dragstart.x() )
	    dragging_adj = -dragging_adj;
	dragging->move(e->x()-dragstart.x(),0);
	layoutTiles();
    }
}

void Rack::contentsMouseReleaseEvent(QMouseEvent* e)
{
    if ( computerized() )
	return;
    if ( dragging ) {
	dragging=0;
	layoutTiles();
    }
}

Bag::Bag()
{
    tiles.setAutoDelete(TRUE);
}

void Bag::add(const Tile& t)
{
    tiles.append(new Tile(t));
}

Tile Bag::takeRandom()
{
    Tile* rp = tiles.take(random()%tiles.count());
    Tile r=*rp;
    return r;
}

ScoreInfo::ScoreInfo( QWidget* parent, const char* name, WFlags fl ) :
    QLabel("<P>",parent,name,fl)
{
    score=0;
    msgtimer = new QTimer(this);
    connect(msgtimer, SIGNAL(timeout()), this, SLOT(showScores()));
}

ScoreInfo::~ScoreInfo()
{
    if ( score ) delete [] score;
}

QSize ScoreInfo::sizeHint() const
{
    return QSize(INT_MAX, fontMetrics().height());
}

void ScoreInfo::init(const QStringList& namelist)
{
    names = namelist;
    if ( score ) delete [] score;
    score = new int[names.count()];
    memset(score,0,sizeof(int)*names.count());
    boldone = -1;
    showScores();
}

void ScoreInfo::addScore(int player, int change)
{
    score[player] += change;
    showScores();
}

void ScoreInfo::setBoldOne(int b)
{
    boldone=b;
    showScores();
}

void ScoreInfo::showScores()
{
    QString r="<p>";
    int i=0;
    int spl=(names.count()+1)/2; // 2 lines
    for (QStringList::ConstIterator it=names.begin(); it!=names.end(); ) {
	if ( i==boldone ) r += "<b>";
	QString n = *it;
	n.replace(QRegExp(":.*"),"");
	r += n;
	r += ":";
	r += QString::number(score[i]);
	if ( i==boldone ) r += "</b>";

	++i;
	++it;
	if ( it != names.end() )
	    r += " ";
    }
    setText(r);
}

void ScoreInfo::showTemporaryScore(int amount)
{
    setText("<P>Score: "+QString::number(amount));
    msgtimer->start(3000,TRUE);
}

